# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.operations.server.shell import Shell
from reemote.execute import execute


def parse_rpm_list_installed(output):
    """Parses the raw string output of `rpm -qa` into a structured list.

    This function processes the multiline string returned by the `rpm -qa`
    command, which lists all installed RPM packages. It intelligently
    separates each line into a package name and its corresponding version string.

    The parsing logic is designed to handle package names that may contain
    hyphens by identifying the version string, which typically starts
    just before the architecture part of the full package identifier.

    Args:
        output (str): The raw stdout from the `rpm -qa` command.

    Returns:
        list[dict]: A list of dictionaries, where each dictionary
            represents an installed package and contains 'name' and
            'version' keys.
    """
    packages = []
    lines = output.strip().splitlines()

    # Skip header line(s) - look for first line that doesn't start with "Installed"
    for line in lines:
        # Split the string by hyphens
        parts = line.split('-')

        # The version starts after the last numeric part before the architecture
        # Find the index where the version starts
        for i in range(len(parts) - 1, 0, -1):
            if parts[i].replace('.', '').isdigit() or parts[i].startswith('p'):
                break

        # Reconstruct the name by joining all parts before the version
        name = '-'.join(parts[:i])

        # Reconstruct the version by joining all parts from the version onward
        version = '-'.join(parts[i:])

        packages.append({"name": name, "version": version})

    return packages

class Get_packages:
    """
    Returns a dictionary of installed packages.

    **Examples:**

    .. code:: python

        yield Get_packages()

    """
    def execute(self):
        from reemote.operations.server.shell import Shell
        r = yield Shell("rpm -qa")
        r.cp.stdout = parse_rpm_list_installed(r.cp.stdout)
        # print(r.cp.stdout)



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command
import base64


class Slurp:
    """
    A class to encapsulate the functionality of the ansible.builtin.slurp module.
    It fetches a base64-encoded blob containing the data in a remote file.

    This module works like ansible.builtin.fetch but returns base64-encoded content.
    It is used for fetching files from remote nodes and is supported for both
    Unix-like and Windows targets.

    Attributes:
        src (str): The file on the remote system to fetch. This must be a file, not a directory.
        guard (bool): If `False` the commands will not be executed.

    **Examples:**

    .. code:: python

        # Fetch a file from remote nodes
        r = yield Slurp(src="/proc/mounts")
        # The result contains base64-encoded content
        print(r.cp.content)

        # Decode the content
        import base64
        decoded_content = base64.b64decode(r.cp.content).decode('utf-8')
        print(decoded_content)

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        - This module returns an 'in memory' base64 encoded version of the file
        - This will require at least twice the RAM as the original file size
        - The src parameter is required and must point to a file, not a directory

    """

    def __init__(self,
                 src: str,
                 guard: bool = True):
        """
        Initialize the Slurp command.

        Args:
            src (str): The file on the remote system to fetch. Required.
            guard (bool): If `False` the commands will not be executed. Defaults to True.
        """
        self.src = src
        self.guard = guard

    def __repr__(self):
        return f"Slurp(src={self.src!r}, guard={self.guard!r})"

    def execute(self):
        """
        Execute the slurp command by reading the file and base64 encoding its content.

        Returns:
            Generator yielding a Command object that reads and encodes the file.
        """
        # Construct command to read file and encode it as base64
        cmd = f"cat {self.src} | base64 -w 0"

        # Execute the command
        r = yield Command(cmd, guard=self.guard)

        # Set the return values to match ansible's slurp module
        if hasattr(r, 'cp') and hasattr(r.cp, 'stdout'):
            r.cp.content = r.cp.stdout.strip()
            r.cp.encoding = "base64"
            r.cp.source = self.src
            r.changed = False  # slurp doesn't change the remote file

        return r


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from typing import Dict, Optional, Callable

class Command:
    """
    Represents a command to be executed on a remote or local host.

    This class encapsulates all the information needed to execute a command,
    including the command string, execution options, and contextual information.

    Attributes:
        command (str): The actual command string to be executed.
        guard (bool): If True, enables guard mode which may prevent execution
                     under certain conditions. Defaults to True.
        host_info (Optional[Dict[str, str]]): Host-specific information and variables.
        global_info (Optional[Dict[str, str]]): Global information and variables
                                               accessible across all hosts.
        local (bool): If True, the command will be executed locally instead of
                     on a remote host. Defaults to False.
        callback (Optional[Callable]): A function to be called after command execution.
        caller: Reference to the entity that initiated the command.
        sudo (bool): If True, the command will be executed with sudo privileges.
                    Defaults to False.
        su (bool): If True, the command will be executed with su privileges.
                  Defaults to False.
        composite (bool): If True, indicates this is a composite command made up
                         of multiple sub-commands. Defaults to False.
    """

    def __init__(self, command: str, guard: bool = True, local: bool = False, callback: Optional[Callable] = None, caller=None, sudo: bool = False, su: bool = False, composite=False):
        self.command: str = command
        self.guard: bool = guard
        self.host_info: Optional[Dict[str, str]] = None
        self.global_info: Optional[Dict[str, str]] = None
        self.local=local
        self.callback=callback
        self.caller=caller
        self.sudo=sudo
        self.su=su
        self.composite=composite


    def __init__(self, command: str, guard: bool = True, local: bool = False, callback: Optional[Callable] = None, caller=None, sudo: bool = False, su: bool = False, composite=False):
        self.command: str = command
        self.guard: bool = guard
        self.host_info: Optional[Dict[str, str]] = None
        self.global_info: Optional[Dict[str, str]] = None
        self.local=local
        self.callback=callback
        self.caller=caller
        self.sudo=sudo
        self.su=su
        self.composite=composite

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return (f"Operation(command={self.command!r}, "
                f"guard={self.guard!r}, "
                f"local={self.local!r}, "
                f"callback={self.callback!r}, "
                f"caller={self.caller!r}, "
                f"composite={self.composite!r}, "
                f"sudo={self.sudo!r}, su={self.su!r}), "
                f"host_info={self.host_info!r}, "
                f"global_info={self.global_info!r})")


def word_wrap(text, width=40):
    wrapped_lines = []

    # Split the text into lines based on existing newlines
    for line in text.splitlines():
        # Handle each line separately
        current_line = ""
        words = line.split()

        for word in words:
            # Check if adding the next word exceeds the width
            if len(current_line) + len(word) + 1 <= width:
                # Add the word to the current line (with a space if not empty)
                current_line += (" " + word if current_line else word)
            else:
                # Append the current line to the result and start a new line
                wrapped_lines.append(current_line)
                current_line = word

        # Append any remaining text in the current line
        if current_line:
            wrapped_lines.append(current_line)

    # Join all wrapped lines with newlines and return
    return "\n".join(wrapped_lines)

def serialize_command(obj):
    if isinstance(obj, Command):  # Check if the object is of type Operation
        return {
            "command": word_wrap(obj.command),          # Serialize the command (string)
            "guard": obj.guard,              # Serialize the guard (boolean)
            "host_info": obj.host_info,      # Serialize the host_info (dict or None)
            "global_info": obj.global_info       # Serialize the global_info (dict or None)
        }
    else:
        raise TypeError(f"Object of type {obj.__class__.__name__} is not JSON serializable")


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
def convert_to_aggrid(df):
    """
    Converts a pandas DataFrame into a format suitable for AG-Grid.

    This function takes a pandas DataFrame and processes it to generate
    the necessary `columnDefs` and `rowData` for use with the AG-Grid
    library. It automatically creates column definitions by inspecting
    the DataFrame's columns and data types.

    Key transformations include:

    - Generating human-readable 'headerName' from column names.
    - Enabling sorting, filtering, and resizing for all columns.
    - Assigning a 'numericColumn' type for integer and float columns.
    - Using an 'agCheckboxCellRenderer' for boolean columns.
    - Converting the DataFrame records into a JSON-serializable list of
      dictionaries, replacing NaN values with None.

    Args:
        df (pd.DataFrame): The input pandas DataFrame to be converted.

    Returns:
        tuple: A tuple containing two elements:

            - columnDefs (list[dict]): A list of dictionaries defining the
              properties for each AG-Grid column.
            - rowData (list[dict]): A list of dictionaries, where each
              dictionary represents a row of data from the DataFrame.
    """
    columnDefs = []
    for column in df.columns:
        col_def = {
            'headerName': column.replace('_', ' ').title(),
            'field': column,
            'sortable': True,
            'filter': True,
            'resizable': True
        }

        # Add specific column properties based on data type
        if df[column].dtype in ['int64', 'float64']:
            col_def['type'] = 'numericColumn'
        elif df[column].dtype == 'bool':
            col_def['cellRenderer'] = 'agCheckboxCellRenderer'

        columnDefs.append(col_def)

    # Convert DataFrame to row data
    rowData = df.replace({float('nan'): None}).to_dict('records')

    return columnDefs, rowData



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import platform
from pathlib import Path
from typing import Optional
import os
import threading
from queue import Queue

from nicegui import events, ui


class local_file_picker(ui.dialog):

    def __init__(self, directory: str, *,
                 upper_limit: Optional[str] = ..., multiple: bool = False,
                 show_hidden_files: bool = False, max_files: int = 1000) -> None:
        super().__init__()

        self.path = Path(directory).expanduser()
        if upper_limit is None:
            self.upper_limit = None
        else:
            self.upper_limit = Path(directory if upper_limit == ... else upper_limit).expanduser()
        self.show_hidden_files = show_hidden_files
        self.max_files = max_files
        self.current_files = []
        self.loading = False

        with self, ui.card():
            # Add a loading indicator
            self.loading_indicator = ui.spinner(size='lg')
            self.loading_indicator.visible = False

            self.grid = ui.aggrid({
                'columnDefs': [{'field': 'name', 'headerName': 'File'}],
                'rowSelection': 'multiple' if multiple else 'single',
                'rowModelType': 'clientSide',
            }, html_columns=[0]).classes('w-96').on('cellDoubleClicked', self.handle_double_click)

            with ui.row().classes('w-full justify-end'):
                ui.button('Cancel', on_click=self.close).props('outline')
                ui.button('Ok', on_click=self._handle_ok)

        # Load files in a separate thread to avoid blocking
        self.load_files_async()

    def load_files_async(self):
        """Load files asynchronously to avoid blocking the main thread"""
        self.loading = True
        self.loading_indicator.visible = True
        self.grid.visible = False

        def load_files():
            try:
                files = []
                count = 0

                # Use scandir for efficient directory reading
                with os.scandir(self.path) as entries:
                    for entry in entries:
                        if count >= self.max_files:
                            break

                        if not self.show_hidden_files and entry.name.startswith('.'):
                            continue

                        files.append({
                            'name': f'📁 <strong>{entry.name}</strong>' if entry.is_dir() else entry.name,
                            'path': entry.path,
                            'is_dir': entry.is_dir()
                        })
                        count += 1

                # Sort directories first, then files
                files.sort(key=lambda x: (not x['is_dir'], x['name'].lower()))

                # Add parent directory if applicable
                if (self.upper_limit is None and self.path != self.path.parent) or \
                        (self.upper_limit is not None and self.path != self.upper_limit):
                    files.insert(0, {
                        'name': '📁 <strong>..</strong>',
                        'path': str(self.path.parent),
                        'is_dir': True
                    })

                self.current_files = files

                # Update UI in main thread
                ui.timer(0.1, once=True, callback=self.update_grid_ui)

            except (PermissionError, OSError) as e:
                ui.timer(0.1, once=True, callback=lambda: self.show_error(f"Error reading directory: {e}"))

        # Start loading in background thread
        thread = threading.Thread(target=load_files, daemon=True)
        thread.start()

    def update_grid_ui(self):
        """Update the grid UI after files are loaded"""
        self.grid.options['rowData'] = self.current_files
        self.grid.update()
        self.loading = False
        self.loading_indicator.visible = False
        self.grid.visible = True

    def show_error(self, message):
        """Show error message"""
        self.loading = False
        self.loading_indicator.visible = False
        ui.notify(message)

    def handle_double_click(self, e: events.GenericEventArguments) -> None:
        new_path = Path(e.args['data']['path'])
        if e.args['data'].get('is_dir', False):
            self.path = new_path
            self.load_files_async()
        else:
            self.submit([str(new_path)])

    async def _handle_ok(self):
        rows = await self.grid.get_selected_rows()
        self.submit([r['path'] for r in rows])


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from nicegui import ui

from reemote.gui.functions.get_versions import get_versions
from reemote.gui.functions.install import install
from reemote.gui.functions.remove import remove
from reemote.gui.functions.update import update
from reemote.gui.functions.upgrade import upgrade


def package_manager(tabs, inventory, versions, manager, stdout_report, execution_report):
    with ui.tab_panels(tabs, value='Package Manager').classes('w-full'):
        with ui.tab_panel('Package Manager'):
            with ui.row():
                ui.select(['apk', 'pip', 'apt', 'dpkg', 'dnf', 'yum'], value='apk').bind_value(manager, 'manager')
                ui.markdown("""
                Choose a package manager from the dropdown list.  
                """)
            with ui.row():
                ui.button('Show installed packages', on_click=lambda: get_versions(inventory, versions, manager, stdout_report, execution_report))
                ui.markdown("""
                Show all of the packages installed on each server.  
                """)
            versions.version_report()
            with ui.row():
                ui.markdown("""
                Install or remove a package from all servers in the inventory.  
                """)
                ui.switch('sudo', value=False).bind_value(manager, 'sudo')
                ui.switch('su', value=False).bind_value(manager, 'su')
                ui.input(label='Package').bind_value(manager, 'package')
                ui.button('Install package', on_click=lambda: install(inventory, versions, manager, stdout_report, execution_report))
                ui.button('Remove package', on_click=lambda: remove(inventory, versions, manager, stdout_report, execution_report))
            with ui.row():
                ui.markdown("""
                Update or Upgrade packages on all servers in the inventory.  
                """)
                ui.button('Update', on_click=lambda: update(inventory, versions, manager, stdout_report, execution_report))
                ui.button('Upgrade', on_click=lambda: upgrade(inventory, versions, manager, stdout_report, execution_report))




# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command


class AddHost:
    """
    A class to encapsulate the functionality of adding hosts to the in-memory inventory
    similar to Ansible's ansible.builtin.add_host module.

    This allows dynamic addition of hosts and groups to the inventory during playbook execution,
    making them available for use in subsequent plays.

    Attributes:
        name (str): The hostname/ip of the host to add to the inventory, can include a colon and a port number.
        groups (list): The groups to add the hostname to.
        **kwargs: Additional variables to assign to the host.

    **Examples:**

    .. code:: python

        # Add a host to a group with custom variables
        yield AddHost(name='192.168.1.100', groups=['web_servers'], foo=42)

        # Add a host to multiple groups
        yield AddHost(hostname='192.168.1.101', groups=['group1', 'group2'])

        # Add a host with non-standard port
        yield AddHost(name='192.168.1.102:2222')

        # Add a host with tunnel configuration
        yield AddHost(hostname='192.168.1.103', ansible_host='localhost', ansible_port=2222)

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        - Hosts added will not bypass the --limit from the command line
        - The host is available from hostvars and for delegation as a normal part of the inventory
        - Changed status is provided as it can be useful for tracking inventory changes
    """

    def __init__(self,
                 name: str = None,
                 hostname: str = None,
                 groups=None,
                 group=None,
                 groupname=None,
                 **kwargs):

        # Handle aliases for name parameter
        if name is None and hostname is not None:
            self.name = hostname
        elif name is None and hostname is None:
            raise ValueError("Either 'name' or 'hostname' must be specified")
        else:
            self.name = name

        # Handle aliases for groups parameter
        if groups is None and group is not None:
            self.groups = [group] if isinstance(group, str) else group
        elif groups is None and groupname is not None:
            self.groups = [groupname] if isinstance(groupname, str) else groupname
        elif groups is None:
            self.groups = []
        else:
            self.groups = groups if isinstance(groups, list) else [groups]

        # Store additional variables as attributes
        self.variables = kwargs

    def __repr__(self):
        groups_repr = repr(self.groups)
        vars_repr = ', '.join([f"{k}={v!r}" for k, v in self.variables.items()])
        return (f"AddHost(name={self.name!r}, "
                f"groups={groups_repr}, "
                f"{vars_repr})")

    def execute(self):
        # Create the command string that would represent this operation
        # In a real implementation, this would interact with the inventory system
        cmd_parts = [f"add_host name={self.name}"]

        if self.groups:
            groups_str = ','.join(self.groups)
            cmd_parts.append(f"groups={groups_str}")

        for key, value in self.variables.items():
            cmd_parts.append(f"{key}={value}")

        cmd = ' '.join(cmd_parts)

        # Yield a command object (placeholder for actual inventory manipulation)
        r = yield Command(cmd, guard=True, sudo=False, su=False)
        r.changed = True
        r.add_host_result = {
            'name': self.name,
            'groups': self.groups,
            'variables': self.variables
        }


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
class Get_list:
    """
    Returns a list of lxc containers from the local host.

    Attributes:

    **Examples:**

    .. code:: python

        yield Get_list()

    """
    def __init__(self,
                 sudo: bool = False,
                 su: bool = False):
        self.sudo: bool = sudo
        self.su: bool = su

    def __repr__(self) -> str:
        return (f"Get_list("
                f"sudo={self.sudo!r}, "
                f"su={self.su!r}"
                f")")


    def execute(self):
        from reemote.operations.server.shell import Shell
        yield Shell("lxc list --format csv | cut -d, -f1",sudo=self.sudo,su=self.su)



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
#!/usr/bin/env python3
from local_file_picker import local_file_picker

from nicegui import ui, native


async def pick_file() -> None:
    result = await local_file_picker('~', multiple=True)
    ui.notify(f'You chose {result}')


@ui.page('/')
def index():
    ui.button('Choose builtin', on_click=pick_file, icon='folder')


ui.run(title="File Controller", reload=False, port=native.find_open_port(),
        storage_secret='private key to secure the browser session cookie')


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.operations.server.shell import Shell
from reemote.execute import execute


def parse_dnf_list_installed(output):
    """Parses the output of 'dnf list installed' into a structured list.

    This function processes the raw string output from commands like
    `dnf list installed` or `yum list installed`. It extracts the package
    name and version for each entry, producing a clean, structured list.

    Key transformations include:

    - Skipping header lines (e.g., "Installed Packages") and empty lines.
    - Handling package names that may contain spaces.
    - Identifying the version by locating the repository field, which starts
      with an '@' symbol, to reliably separate it from the name.
    - Assembling a list of dictionaries, each with 'name' and 'version' keys.

    Args:
        output (str): The raw standard output from the dnf/yum command.

    Returns:
        list[dict]: A list of dictionaries, where each dictionary represents
          an installed package and contains 'name' and 'version' keys.
    """
    packages = []
    lines = output.strip().splitlines()

    # Skip header line(s) - look for first line that doesn't start with "Installed"
    start_processing = False
    for line in lines:
        line = line.strip()
        if not line:
            continue
        if line.startswith("Installed Packages"):
            start_processing = True
            continue
        if not start_processing:
            continue

        # Split line by whitespace, but we need to be careful because name and version can have spaces
        # The format is: NAME VERSION REPO
        # We know repo starts with @, so we can find the last field that starts with @
        parts = line.split()
        if len(parts) < 3:
            continue

        # Find the index where the repo (starting with @) begins
        repo_index = -1
        for i in range(len(parts) - 1, -1, -1):
            if parts[i].startswith('@'):
                repo_index = i
                break

        if repo_index == -1:
            continue

        # Version is the part just before the repo
        version = parts[repo_index - 1]
        # Name is everything from the start up to (but not including) the version
        name_parts = parts[:repo_index - 1]
        name = ' '.join(name_parts) if len(name_parts) > 1 else name_parts[0]

        packages.append({"name": name, "version": version})

    return packages

class Get_packages:
    """
    Returns a dictionary of installed packages.

    **Examples:**

    .. code:: python

        yield Get_packages()

    """
    def execute(self):
        from reemote.operations.server.shell import Shell
        r = yield Shell("yum list installed")
        r.cp.stdout = parse_dnf_list_installed(r.cp.stdout)
        # print(r.cp.stdout)



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command

class Shell:
    """
    A class to encapsulate the functionality of shell commands in Unix-like operating systems.
    It allows users to specify a shell command that is executed on all hosts.
    additional command-line options, and the ability to execute the command with elevated privileges (`sudo`).

    Attributes:
        cmd (str): The shell command.
        guard (bool): If `False` the commands will not be executed.
        sudo (bool): If `True`, the commands will be executed with `sudo` privileges.
        su (bool): If `True`, the commands will be executed with `su` privileges.

    **Examples:**

    .. code:: python

        # Execute a shell command on all hosts
        r = yield Shell("echo Hello World")
        # The result is available in stdout
        print(r.cp.stdout)

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        - Commands are constructed based on the `sudo`, and `su` flags.

    """
    def __init__(self,
                 cmd: str,
                 guard: bool = True,
                 sudo: bool = False,
                 su: bool = False):

        self.cmd = cmd
        self.guard = guard
        self.sudo = sudo
        self.su = su

    def __repr__(self):
        return (f"Shell(cmd={self.cmd!r}, "
                f"guard={self.guard!r}, "
                f"sudo={self.sudo!r}, su={self.su!r})")

    def execute(self):
        # print(f"{self}")
        r = yield Command(self.cmd, guard=self.guard, sudo=self.sudo, su=self.su)
        r.changed = True



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
class Operation_upgrade:
    """
    Manages update operations under specific guard and privilege settings.

    This class is responsible for performing updates with options to
    guard the operation, run as superuser using 'sudo', or switch user
    using 'su'. It provides an interface for controlled execution of
    update commands.

    Attributes:
        guard: Specifies whether to enable guard to monitor changes during
               the update process (default is True).
        sudo: A flag to indicate if the update should be executed with
              superuser privileges using 'sudo' (default is False).
        su: A flag to indicate if the update should be executed with
            a user switch (default is False).

    Methods:
        execute:
            Executes the update process while optionally monitoring for
            changes and employing privilege settings depending on `guard`,
            `sudo`, and `su` attributes.
    """

    def __init__(self,
                 guard: bool = True,
                 sudo: bool = False,
                 su: bool = False):
        self.guard = guard
        self.sudo: bool = sudo
        self.su: bool = su

    def __repr__(self) -> str:
        return (f"Upgrade("
                f"guard={self.guard!r}, "                
                f"sudo={self.sudo!r}, "
                f"su={self.su!r})")

    def execute(self):
        r1 = yield self.get_packages()
        r2 = yield self.upgrade()
        r3 = yield self.get_packages()
        # Set the `changed` flag if the package state has changed
        if self.guard and (r1.cp.stdout != r3.cp.stdout):
            r2.changed = True

    def get_packages(self):
        """
        Retrieve the list of installed packages.
        Subclasses must implement this method.
        """
        raise NotImplementedError("Subclasses must implement `get_packages`.")

    def upgrade(self,guard=None,sudo=None,su=None):
        """
        Install the specified packages.
        Subclasses must implement this method.
        """
        raise NotImplementedError("Subclasses must implement `upgrade`.")



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import argparse
import asyncio
import sys
from reemote.execute import execute
from reemote.utilities.produce_table import produce_table
from reemote.utilities.produce_output_table import produce_output_table
from reemote.utilities.produce_json import produce_json
from reemote.utilities.verify_python_file import verify_python_file
from reemote.utilities.validate_inventory_file_and_get_inventory import validate_inventory_file_and_get_inventory
from reemote.utilities.validate_inventory_structure import validate_inventory_structure
from reemote.operations.filesystem.directory import Directory

def str_to_bool(value):
    """
    Convert a string representation of a boolean to a Python boolean.
    Accepts 'True', 'true', 'False', 'false' as valid inputs.
    Raises an error for invalid inputs.
    """
    value = value.lower()
    if value in ('true', 't'):
        return True
    elif value in ('false', 'f'):
        return False
    else:
        raise argparse.ArgumentTypeError(f"Invalid boolean value: {value}")

async def main():
    # Create an ArgumentParser object
    parser = argparse.ArgumentParser(description="Process inventory and directory information asynchronously.")

    # Add arguments
    parser.add_argument(
        "-i", "--inventory",
        type=str,
        required=True,
        help="Path to the inventory builtin (must exist)."
    )
    parser.add_argument(
        "-d", "--directory",
        type=str,
        required=True,
        help="Path to the directory."
    )
    parser.add_argument(
        "-p", "--present",
        type=str_to_bool,  # Custom function to convert string to boolean
        required=True,
        help="Boolean flag (True or False)."
    )

    # Parse the arguments
    args = parser.parse_args()

    # Print the parsed arguments for verification
    print("Inventory File:", args.inventory)
    print("Directory:", args.directory)
    print("Present Flag:", args.present)



    # Verify inventory builtin
    if args.inventory:
        if not verify_python_file(args.inventory):
            sys.exit(1)

    # verify the inventory
    if args.inventory:
        inventory = validate_inventory_file_and_get_inventory(args.inventory)
        if not inventory:
            sys.exit(1)
    else:
        inventory = []

    if args.inventory:
        if not validate_inventory_structure(inventory()):
            print("Inventory structure is invalid")
            sys.exit(1)

    responses = await execute(inventory(), Directory(path=args.directory, present=args.present))
    print(produce_table(produce_json(responses)))
    print(produce_output_table(produce_json(responses)))


if __name__ == "__main__":
    # Run the asyncio event loop
    asyncio.run(main())



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command


class RpmKey:
    """
    A class to manage GPG keys in the RPM database on RPM-based systems.
    This class provides functionality to add or remove GPG keys from the RPM database,
    supporting various key sources including URLs, local files, and existing key IDs.

    Attributes:
        key (str): Key that will be modified. Can be a url, a file on the managed node, or a keyid if the key already exists in the database.
        state (str): If the key will be imported or removed from the rpm db. Choices: "absent", "present".
        fingerprint (list): The long-form fingerprint(s) of the key being imported for verification.
        validate_certs (bool): If false and the key is a url starting with https, SSL certificates will not be validated.

    **Examples:**

    .. code:: python

        # Import a key from a url
        r = yield RpmKey(state="present", key="http://apt.sw.be/RPM-GPG-KEY.dag.txt")

        # Import a key from a file
        r = yield RpmKey(state="present", key="/path/to/key.gpg")

        # Ensure a key is not present in the db
        r = yield RpmKey(state="absent", key="DEADB33F")

        # Verify the key, using a fingerprint, before import
        r = yield RpmKey(key="/path/to/RPM-GPG-KEY.dag.txt", fingerprint="EBC6 E12C 62B1 C734 026B  2122 A20E 5214 6B8D 79E6")

        # Verify the key, using multiple fingerprints, before import
        r = yield RpmKey(key="/path/to/RPM-GPG-KEY.dag.txt", fingerprint=["EBC6 E12C 62B1 C734 026B  2122 A20E 5214 6B8D 79E6", "19B7 913E 6284 8E3F 4D78 D6B4 ECD9 1AB2 2EB6 8D86"])

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        - Commands are constructed based on the provided parameters and state.
        - Supports both adding (present) and removing (absent) keys from the RPM database.
        - Can verify keys using fingerprints before importing.
    """

    def __init__(self,
                 key: str,
                 state: str = "present",
                 fingerprint: list = None,
                 validate_certs: bool = True):

        self.key = key
        self.state = state
        self.fingerprint = fingerprint if fingerprint is not None else []
        self.validate_certs = validate_certs

        # Normalize fingerprint to list if single string provided
        if isinstance(self.fingerprint, str):
            self.fingerprint = [self.fingerprint]

    def __repr__(self):
        return (f"RpmKey(key={self.key!r}, "
                f"state={self.state!r}, "
                f"fingerprint={self.fingerprint!r}, "
                f"validate_certs={self.validate_certs!r})")

    def execute(self):
        # Build the rpm command based on state
        if self.state == "present":
            cmd_parts = ["rpm", "--import"]

            # Add key verification if fingerprints are provided
            if self.fingerprint:
                # For verification, we would typically check the key first
                # This is a simplified approach - in practice you might want to verify before import
                pass

            cmd_parts.append(self.key)
            cmd = " ".join(cmd_parts)

        elif self.state == "absent":
            # To remove a key, we need to find it first and then remove it
            # This is more complex and typically involves querying the rpmdb
            # For simplicity, we'll use rpm -e with the key ID
            cmd = f"rpm -e gpg-pubkey-{self.key.lower()}* 2>/dev/null || true"
        else:
            raise ValueError(f"Invalid state '{self.state}'. Must be 'present' or 'absent'.")

        # Execute the command
        r = yield Command(cmd)
        r.changed = True


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import asyncssh
from reemote.command import Command
from typing import Optional, Callable, Union


class Copy_files:
    """
    A class to encapsulate the functionality of remote-to-remote builtin copying using SFTP.
    It allows users to copy multiple remote builtin to new remote locations with full parameter support.

    Attributes:
        srcpaths (str): The remote source builtin or directory path(s) to copy.
        dstpath (str): The remote destination path where builtin will be copied.
        preserve (bool): Preserve builtin attributes (permissions, timestamps).
        recurse (bool): Recursively copy directories.
        follow_symlinks (bool): Follow symbolic links during copy.
        sparse (bool): Create sparse builtin on the remote system.
        block_size (int): Block size for builtin transfers.
        max_requests (int): Maximum number of concurrent transfer requests.
        progress_handler (Callable): Callback for copy progress.
        error_handler (Callable): Callback for handling errors.
        remote_only (bool): Whether to only allow remote copy operations.

    **Examples:**

    .. code:: python

        src_dir = '/home/user/'
        dst_dir = '/home/user/'
        r = yield Copy_files(
            srcpaths=src_dir + '/example.txt',
            dstpath=dst_dir+ '/example1.txt',
            preserve=True,
            recurse=True,
            progress_handler=my_progress_callback,
        )

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.
    """

    def __init__(self,
                 srcpaths: str,
                 dstpath: str,
                 preserve: bool = False,
                 recurse: bool = False,
                 follow_symlinks: bool = False,
                 sparse: bool = True,
                 block_size: int = -1,
                 max_requests: int = -1,
                 progress_handler: Optional[Callable] = None,
                 error_handler: Optional[Callable] = None,
                 remote_only: bool = True):
        self.srcpaths = srcpaths
        self.dstpath = dstpath
        self.preserve = preserve
        self.recurse = recurse
        self.follow_symlinks = follow_symlinks
        self.sparse = sparse
        self.block_size = block_size
        self.max_requests = max_requests
        self.progress_handler = progress_handler
        self.error_handler = error_handler
        self.remote_only = remote_only

    def __repr__(self):
        return (f"Copy_files(srcpaths={self.srcpaths!r}, "
                f"dstpath={self.dstpath!r}, "
                f"preserve={self.preserve!r}, "
                f"recurse={self.recurse!r}, "
                f"follow_symlinks={self.follow_symlinks!r}, "
                f"sparse={self.sparse!r}, "
                f"block_size={self.block_size!r}, "
                f"max_requests={self.max_requests!r}, "
                f"remote_only={self.remote_only!r})")

    @staticmethod
    async def _copy_callback(host_info, global_info, command, cp, caller):
        """Static callback method for remote-to-remote builtin copying with full parameter support"""
        
        # Validate host_info
        required_keys = ['host', 'username', 'password']
        for key in required_keys:
            if key not in host_info or host_info[key] is None:
                raise ValueError(f"Missing or invalid value for '{key}' in host_info.")

        # Validate caller attributes
        if caller.srcpaths is None:
            raise ValueError("The 'srcpaths' attribute of the caller cannot be None.")
        if caller.dstpath is None:
            raise ValueError("The 'dstpath' attribute of the caller cannot be None.")

        try:
            # Connect to the SSH server
            async with asyncssh.connect(**host_info) as conn:
                # Start an SFTP session
                async with conn.start_sftp_client() as sftp:
                    # Use the copy method for remote-to-remote copying
                    await sftp.copy(
                        srcpaths=caller.srcpaths,
                        dstpath=caller.dstpath,
                        preserve=caller.preserve,
                        recurse=caller.recurse,
                        follow_symlinks=caller.follow_symlinks,
                        sparse=caller.sparse,
                        block_size=caller.block_size,
                        max_requests=caller.max_requests,
                        progress_handler=caller.progress_handler,
                        error_handler=caller.error_handler,
                        remote_only=caller.remote_only
                    )
                    return f"Successfully copied builtin on {host_info['host']}"
        except (OSError, asyncssh.Error) as exc:
            raise  # Re-raise the exception to handle it in the caller

    def execute(self):
        r = yield Command(f"{self}", local=True, callback=self._copy_callback, caller=self)
        r.executed = True
        r.changed = True
        return r



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from typing import List
from reemote.operation_upgrade import Operation_upgrade
from reemote.commands.yum.upgrade import Upgrade
from reemote.facts.yum.get_packages import Get_packages


class Upgrade(Operation_upgrade):
    """
    A class to manage package upgrade operations on a remote system using `yum`.

    This class provides functionality to upgrade installed packages on RHEL/CentOS systems
    using the yum package manager. It allows configuration of execution privileges and
    safety guards for package upgrade operations.

    Attributes:
        guard (bool): If `False` the commands will not be executed.
        sudo (bool): If `True`, the commands will be executed with sudo privileges.
        su (bool): If `True`, the commands will be executed with su privileges.

    **Examples:**

    .. code:: python

        # Upgrade installed packages on all hosts
        r = yield Upgrade()
        # Check if the operation was successful
        if r.cp.return_code == 0:
            print("Packages upgraded successfully")

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        - Uses `yum upgrade` command internally for upgrading packages.
        - Inherits from Operation_upgrade base class.
        - This operation upgrades installed packages to their latest versions.
    """


    def __init__(self,
                 guard: bool = True,
                 sudo: bool = False,
                 su: bool = False):
        super().__init__(guard, sudo, su)

    def get_packages(self):
        return Get_packages()

    def upgrade_packages(self, guard=None,sudo=None,su=None):
        from reemote.commands.yum.upgrade import Upgrade
        return Upgrade(self.guard, self.sudo, self.su)



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from typing import List
from reemote.command import Command
from reemote.utilities.validate_list_of_strings import validate_list_of_strings
class Command_install:
    """
    Represents an installation or removal operation for packages.

    This class is designed to manage package installations with options for adding
    guards, sudo privileges, or the su user. It constructs the operation string
    from the provided packages and represents an encapsulated way of handling
    package installation commands.

    Attributes:
        packages: List of package names to be installed.
        guard: A boolean flag indicating whether the operation should be guarded.
        sudo: A boolean flag to specify if sudo privileges are required.
        su: A boolean flag to specify if the operation should run as su.

    """
    def __init__(self,
                 packages: List[str],
                 guard: bool = True,
                 sudo: bool = False,
                 su: bool = False):

        self.packages: List[str] = packages
        self.guard: bool = guard
        self.sudo: bool = sudo
        self.su: bool = su

        # Ensure self.packages is always a list of strings
        self.packages: List[str] = validate_list_of_strings(packages)

        # Construct the operation string from the list of packages
        op: List[str] = []
        op.extend(self.packages)
        self.op: str = " ".join(op)

    def __repr__(self) -> str:
        return (f"Command("
                f"packages={self.packages!r}, "
                f"guard={self.guard!r}, "                                
                f"sudo={self.sudo!r}, "
                f"su={self.su!r})")





# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import asyncssh
from reemote.command import Command


class Chdir:
    """
    A class to encapsulate the functionality of chdir (change directory) in
    Unix-like operating systems.

    Attributes:
        path (str): The directory path to change to.

    **Examples:**

    .. code:: python

        yield Chdir(path='/home/user/hfs')

    Usage:
        This class is designed to be used in a generator-based workflow where
        commands are yielded for execution.
    """

    def __init__(self, path: str):
        self.path = path

    def __repr__(self):
        return f"Chdir(path={self.path!r})"

    @staticmethod
    async def _chdir_callback(host_info, global_info, command, cp, caller):
        """Static callback method for directory change"""

        # Validate host_info
        required_keys = ['host', 'username', 'password']
        for key in required_keys:
            if key not in host_info or host_info[key] is None:
                raise ValueError(f"Missing or invalid value for '{key}' in host_info.")

        # Validate caller attributes
        if caller.path is None:
            raise ValueError("The 'path' attribute of the caller cannot be None.")

        try:
            async with asyncssh.connect(**host_info) as conn:
                async with conn.start_sftp_client() as sftp:
                    # Change the current remote working directory
                    await sftp.chdir(path=caller.path)
                    return f"Successfully changed directory to {caller.path} on {host_info['host']}"
        except (OSError, asyncssh.Error) as exc:
            raise  # Re-raise the exception to handle it in the caller

    def execute(self):
        r = yield Command(f"{self}", local=True, callback=self._chdir_callback, caller=self)
        r.executed = True
        r.changed = False
        return r



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import importlib.util
import sys
from typing import Any


def validate_root_class_name_and_get_root_class(class_name, source_file) -> Any:
    """Dynamically loads a class from a Python source builtin and validates its existence.

    This function uses the `importlib` utility to load a Python source builtin
    as a module at runtime. It creates a module from the specified builtin,
    executes its content, and then attempts to locate a class with the given
    `class_name` within that module.

    The process involves:

    - Creating a module specification from the builtin path.
    - Executing the module's code to populate it with functions and classes.
    - Optionally adding the new module to `sys.modules` for broader accessibility.
    - Checking for the presence of the target class. If the class is not
      found, an error message is printed to standard output.

    Args:
        class_name (str): The name of the class to validate and retrieve from
            the source builtin.
        source_file (str): The builtin path to the Python source builtin (.py)
            containing the class definition.

    Returns:
        typing.Any: The class object if it is found within the source builtin.
        Returns `False` if the class with the specified `class_name`
        does not exist in the builtin.
    """
    module_name = "dynamic_module"  # You can name this anything
    spec = importlib.util.spec_from_file_location(module_name, source_file)
    # Create a new module based on the specification
    module = importlib.util.module_from_spec(spec)
    # Execute the module (this runs the code in the builtin)
    spec.loader.exec_module(module)

    # Optionally, add the module to sys.modules so it behaves like a regular import
    sys.modules[module_name] = module

    # Now you can access functions and classes defined in the builtin
    # Example:
    if not hasattr(module, class_name):
        print(f"Source builtin must contain class {class_name}")
        return False
    else:
        # Access the `inventory` function from the module
        root_class = getattr(module, class_name)
    return root_class



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import asyncssh
from reemote.command import Command


class Mkdir:
    """
    A class to encapsulate the functionality of mkdir in Unix-like operating systems.

    Attributes:
        path (str): The directory path to create.

    **Examples:**

    .. code:: python

        yield Mkdir(path='/home/user/hfs')

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.
    """

    def __init__(self, path: str):
        self.path = path

    def __repr__(self):
        return f"Mkdir(path={self.path!r})"

    @staticmethod
    async def _mkdir_callback(host_info, global_info, command, cp, caller):
        """Static callback method for directory creation"""

        # Validate host_info with proper error messages
        required_keys = ['host', 'username', 'password']
        missing_keys = []
        invalid_keys = []

        for key in required_keys:
            if key not in host_info:
                missing_keys.append(key)
            elif host_info[key] is None:
                invalid_keys.append(key)

        if missing_keys:
            raise ValueError(f"Missing required keys in host_info: {missing_keys}")
        if invalid_keys:
            raise ValueError(f"None values for keys in host_info: {invalid_keys}")

        # Validate caller attributes
        if caller.path is None:
            raise ValueError("The 'path' attribute of the caller cannot be None.")

        # Clean host_info by removing None values and keys that asyncssh doesn't expect
        clean_host_info = {
            'host': host_info['host'],
            'username': host_info['username'],
            'password': host_info['password']
        }

        # Add optional parameters only if they exist and are not None
        optional_keys = ['port', 'known_hosts', 'client_keys', 'passphrase']
        for key in optional_keys:
            if key in host_info and host_info[key] is not None:
                clean_host_info[key] = host_info[key]

        try:
            async with asyncssh.connect(**clean_host_info) as conn:
                async with conn.start_sftp_client() as sftp:
                    # Create the remote directory
                    await sftp.mkdir(caller.path)
                    return f"Successfully created directory {caller.path} on {host_info['host']}"

        except (OSError, asyncssh.Error) as exc:
            # Provide more detailed error information
            raise Exception(f"Failed to create directory {caller.path} on {host_info['host']}: {str(exc)}")

    def execute(self):
        r = yield Command(f"{self}", local=True, callback=self._mkdir_callback, caller=self)
        r.executed = True
        # Directory creation is inherently a changing operation
        r.changed = True
        return r



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import asyncssh
from asyncssh.sftp import SFTPAttrs
class Setstat:
    def __init__(self, path: str, attrs, follow_symlinks: bool = True):
        self.path = path
        self.attrs = attrs
        self.follow_symlinks = follow_symlinks

        # Convert attrs to a dictionary
        if isinstance(self.attrs, asyncssh.SFTPAttrs):
            # Manually construct a dictionary from SFTPAttrs, filtering out None values
            self.attrs_dict = {
                key: value
                for key, value in self.attrs.__dict__.items()
                if value is not None
            }
        elif isinstance(self.attrs, dict):
            # Use the existing dictionary
            self.attrs_dict = self.attrs
        else:
            raise ValueError("The 'attrs' parameter must be an SFTPAttrs object or a dictionary.")

        # Ensure the dictionary is non-empty
        if not self.attrs_dict:
            raise ValueError("The 'attrs' attribute must be a non-empty dictionary.")

        # Debug: Inspect the attrs_dict
        print(f"Converted attrs to dictionary: {self.attrs_dict}")

    @staticmethod
    async def _set_attributes_callback(host_info, global_info, command, cp, caller):
        """Static callback method to set builtin attributes using SFTP."""

        print("Executing _set_attributes_callback...")  # Debug statement

        # Validate host_info
        required_keys = ['host', 'username', 'password']
        for key in required_keys:
            if key not in host_info or host_info[key] is None:
                raise ValueError(f"Missing or invalid value for '{key}' in host_info.")

        # Validate caller attributes
        if caller.path is None:
            raise ValueError("The 'path' attribute of the caller cannot be None.")

        try:
            async with asyncssh.connect(**host_info) as conn:
                async with conn.start_sftp_client() as sftp:
                    # Check if the builtin exists
                    try:
                        await sftp.stat(caller.path)
                    except FileNotFoundError:
                        raise FileNotFoundError(f"File {caller.path} does not exist on {host_info['host']}.")

                    # Set the builtin attributes using setstat
                    await sftp.setstat(caller.path, caller.attrs_dict, follow_symlinks=caller.follow_symlinks)

                    # Build result message
                    result_msg = f"Set attributes for builtin {caller.path} on {host_info['host']}"

                    # Add attribute information to result message
                    attr_details = []
                    if caller.attrs_dict.get('permissions') is not None:
                        attr_details.append(f"permissions={oct(caller.attrs_dict['permissions'])}")
                    if caller.attrs_dict.get('uid') is not None:
                        attr_details.append(f"uid={caller.attrs_dict['uid']}")
                    if caller.attrs_dict.get('gid') is not None:
                        attr_details.append(f"gid={caller.attrs_dict['gid']}")

                    if attr_details:
                        result_msg += f" with attributes: {', '.join(attr_details)}"

                    return result_msg

        except (OSError, asyncssh.Error) as exc:
            raise  # Re-raise the exception to handle it in the caller

    def execute(self):
        r = yield Command(f"{self}", local=True, callback=self._set_attributes_callback, caller=self)
        r.executed = True
        r.changed = True  # Set to True since setstat changes the system state
        return r


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command


class AsyncStatus:
    """
    A class to encapsulate the functionality of the ansible.builtin.async_status module.
    This module gets the status of an asynchronous task and can also clean up the async job cache.

    Attributes:
        jid (str): Job or task identifier (required).
        mode (str): If "status", obtain the status. If "cleanup", clean up the async job cache.
                   Defaults to "status".

    **Examples:**

    .. code:: python

        # Wait for asynchronous job to end
        r = yield AsyncStatus(jid=dnf_sleeper.ansible_job_id)
        # Check if job is finished
        if r.cp.finished:
            print("Job completed")

        # Clean up async file
        r = yield AsyncStatus(jid=dnf_sleeper.ansible_job_id, mode="cleanup")

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        - The started and finished return values return True or False instead of 1 or 0.
        - This module is supported for both POSIX and Windows targets.
        - The module has a corresponding action plugin.
    """

    def __init__(self,
                 jid: str,
                 mode: str = "status"):
        """
        Initialize the AsyncStatus object.

        Args:
            jid (str): Job or task identifier (required).
            mode (str): Operation mode - either "status" or "cleanup". Defaults to "status".
        """
        self.jid = jid
        self.mode = mode

        # Validate mode parameter
        if self.mode not in ["status", "cleanup"]:
            raise ValueError("mode must be either 'status' or 'cleanup'")

    def __repr__(self):
        return f"AsyncStatus(jid={self.jid!r}, mode={self.mode!r})"

    def execute(self):
        """
        Execute the async_status command.

        Returns:
            Generator yielding Command object for execution.
        """
        # Construct the ansible command
        if self.mode == "cleanup":
            cmd = f"ansible.builtin.async_status jid={self.jid} mode=cleanup"
        else:
            cmd = f"ansible.builtin.async_status jid={self.jid}"

        r = yield Command(cmd)
        r.changed = True


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import importlib.util
import inspect


def verify_source_file_contains_valid_class(source_file, class_name):
    """Verifies a Python source builtin contains a specific class with a valid `execute` method.

     This function dynamically loads a Python module from a given source builtin path
     and performs a series of validations to ensure it conforms to a specific
     interface. It is designed to check for "plugin" or "task" style classes
     before they are executed.

     The validation steps are as follows:

     - It attempts to load the module from the `source_file`.
     - It checks if a class with the name `class_name` exists in the module.
     - It verifies that this class contains a method named `execute`.
     - It inspects the signature of the `execute` method to confirm it accepts
       no arguments, other than the implicit `self`.

     If any of these checks fail or if an exception occurs during module loading,
     an error message is printed to the console, and the function returns `False`.

     Args:
         source_file (str): The builtin path of the Python source code to inspect.
         class_name (str): The name of the class to validate within the source builtin.

     Returns:
         bool: `True` if the source builtin contains the specified class with a
               valid `execute` method, `False` otherwise.
    """
    try:
        # Load the module from builtin
        spec = importlib.util.spec_from_file_location("source_module", source_file)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        # Check if class exists
        if not hasattr(module, class_name):
            print(f"Error: Class '{class_name}' not found in {source_file}")
            return False

        cls = getattr(module, class_name)

        # Check if execute method exists
        if not hasattr(cls, 'execute'):
            print(f"Error: Class '{class_name}' does not have an execute method")
            return False

        # Check if execute method takes no parameters (besides self)
        execute_method = getattr(cls, 'execute')
        sig = inspect.signature(execute_method)

        # Should have only 'self' parameter
        if len(sig.parameters) != 1:
            print(
                f"Error: execute method should take only 'self' parameter, but takes {len(sig.parameters)} parameters")
            return False

        return True

    except Exception as e:
        print(f"Error loading module from {source_file}: {e}")
        return False



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import time


class Rate_limiting_progress_updates:
    """
    Progress handler that throttles updates to avoid flooding

    Args:
        src_path: Source builtin path
        dst_path: Destination builtin path
        copied_bytes: Number of bytes copied so far
        total_bytes: Total bytes to copy (None if unknown)
    """
    def __init__(self, min_interval=0.1):  # Update at most every 100ms
        self.min_interval = min_interval
        self.last_update = 0

    def __call__(self, src_path, dst_path, copied_bytes, total_bytes):
        current_time = time.time()
        if current_time - self.last_update >= self.min_interval:
            if total_bytes:
                percentage = (copied_bytes / total_bytes) * 100
                print(f"Progress: {percentage:.1f}% ({copied_bytes}/{total_bytes} bytes)")
            else:
                print(f"Progress: {copied_bytes} bytes copied")
            self.last_update = current_time





# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
class Exec:
    """
    A class to manage lxc operations from the local host.

    Attributes:
        vm (str): Name of the virtual machine
        cmd (str): Command to execute
        sudo (bool): If `True`, the commands will be executed with `sudo` privileges.
        su (bool): If `True`, the commands will be executed with `su` privileges.

    **Examples:**

    .. code:: python

        yield Exec(vm='debian-vm2',cmd='apt-get update')

    Usage:
        Execute command in container.

    Notes:
        - Commands are constructed based on the `present`, `sudo`, and `su` flags.
        - The `changed` flag is set if the package state changes after execution.
    """

    def __init__(self,
                 vm: str,
                 cmd: str,
                 sudo: bool = False,
                 su: bool = False):
        self.cmd = cmd
        self.vm = vm
        self.sudo: bool = sudo
        self.su: bool = su

    def __repr__(self) -> str:
        return (f"Exec("
                f"vm={self.vm!r}, "
                f"cmd={self.cmd!r}, "
                f"sudo={self.sudo!r}, "
                f"su={self.su!r}"
                f")")

    def execute(self):
        from reemote.operations.server.shell import Shell
        yield Shell(f"lxc exec {self.vm} -- {self.cmd}",sudo=self.sudo,su=self.su)



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command


class Cron:
    """
    A class to manage cron.d and crontab entries in Unix-like operating systems.
    This class allows you to create environment variables and named crontab entries,
    update, or delete them.

    When crontab jobs are managed: the module includes one line with the description
    of the crontab entry "#Ansible: <name>" corresponding to the name passed to the
    module, which is used by future ansible/module calls to find/check the state.
    The name parameter should be unique, and changing the name value will result in
    a new cron task being created (or a different one being removed).

    When environment variables are managed, no comment line is added, but, when the
    module needs to find/check the state, it uses the name parameter to find the
    environment variable definition line.

    When using symbols such as %, they must be properly escaped.

    Attributes:
        name (str): Description of a crontab entry or, if env is set, the name of environment variable.
        user (str): The specific user whose crontab should be modified.
        minute (str): Minute when the job should run (0-59, *, */2, and so on).
        hour (str): Hour when the job should run (0-23, *, */2, and so on).
        day (str): Day of the month the job should run (1-31, *, */2, and so on).
        month (str): Month of the year the job should run (JAN-DEC or 1-12, *, */2, and so on).
        weekday (str): Day of the week that the job should run (SUN-SAT or 0-6, *, and so on).
        job (str): The command to execute or, if env is set, the value of environment variable.
        special_time (str): Special time specification nickname.
        disabled (bool): If the job should be disabled (commented out) in the crontab.
        env (bool): If set, manages a crontab's environment variable.
        state (str): Whether to ensure the job or environment variable is present or absent.
        backup (bool): If set, create a backup of the crontab before it is modified.
        cron_file (str): If specified, uses this file instead of an individual user's crontab.
        insertafter (str): Used with state=present and env. Insert after specified environment variable.
        insertbefore (str): Used with state=present and env. Insert before specified environment variable.

    **Examples:**

    .. code:: python

        # Ensure a job that runs at 2 and 5 exists
        r = yield Cron(
            name="check dirs",
            minute="0",
            hour="5,2",
            job="ls -alh > /dev/null"
        )

        # Ensure an old job is no longer present
        r = yield Cron(
            name="an old job",
            state="absent"
        )

        # Creates an entry like "@reboot /some/job.sh"
        r = yield Cron(
            name="a job for reboot",
            special_time="reboot",
            job="/some/job.sh"
        )

        # Creates an entry like "PATH=/opt/bin" on top of crontab
        r = yield Cron(
            name="PATH",
            env=True,
            job="/opt/bin"
        )

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        - If you are experiencing permissions issues with cron and MacOS, you should see the official MacOS documentation for further information.
        - Requires cron (any 'vixie cron' conformant variant, like cronie)
    """

    def __init__(self,
                 name: str,
                 user: str = None,
                 minute: str = "*",
                 hour: str = "*",
                 day: str = "*",
                 month: str = "*",
                 weekday: str = "*",
                 job: str = None,
                 special_time: str = None,
                 disabled: bool = False,
                 env: bool = False,
                 state: str = "present",
                 backup: bool = False,
                 cron_file: str = None,
                 insertafter: str = None,
                 insertbefore: str = None):

        self.name = name
        self.user = user
        self.minute = minute
        self.hour = hour
        self.day = day
        self.month = month
        self.weekday = weekday
        self.job = job
        self.special_time = special_time
        self.disabled = disabled
        self.env = env
        self.state = state
        self.backup = backup
        self.cron_file = cron_file
        self.insertafter = insertafter
        self.insertbefore = insertbefore

    def __repr__(self):
        return (f"Cron(name={self.name!r}, "
                f"user={self.user!r}, "
                f"minute={self.minute!r}, "
                f"hour={self.hour!r}, "
                f"day={self.day!r}, "
                f"month={self.month!r}, "
                f"weekday={self.weekday!r}, "
                f"job={self.job!r}, "
                f"special_time={self.special_time!r}, "
                f"disabled={self.disabled!r}, "
                f"env={self.env!r}, "
                f"state={self.state!r}, "
                f"backup={self.backup!r}, "
                f"cron_file={self.cron_file!r}, "
                f"insertafter={self.insertafter!r}, "
                f"insertbefore={self.insertbefore!r})")

    def execute(self):
        # Build the command arguments
        cmd_args = ["ansible.builtin.cron"]

        # Add required name parameter
        cmd_args.append(f"name={self.name}")

        # Add optional parameters
        if self.user:
            cmd_args.append(f"user={self.user}")
        if self.minute != "*":
            cmd_args.append(f"minute={self.minute}")
        if self.hour != "*":
            cmd_args.append(f"hour={self.hour}")
        if self.day != "*":
            cmd_args.append(f"day={self.day}")
        if self.month != "*":
            cmd_args.append(f"month={self.month}")
        if self.weekday != "*":
            cmd_args.append(f"weekday={self.weekday}")
        if self.job:
            cmd_args.append(f"job={self.job}")
        if self.special_time:
            cmd_args.append(f"special_time={self.special_time}")
        if self.disabled:
            cmd_args.append("disabled=yes")
        if self.env:
            cmd_args.append("env=yes")
        if self.state != "present":
            cmd_args.append(f"state={self.state}")
        if self.backup:
            cmd_args.append("backup=yes")
        if self.cron_file:
            cmd_args.append(f"cron_file={self.cron_file}")
        if self.insertafter:
            cmd_args.append(f"insertafter={self.insertafter}")
        if self.insertbefore:
            cmd_args.append(f"insertbefore={self.insertbefore}")

        # Construct the full command
        cmd = " ".join(cmd_args)

        # Execute the command
        r = yield Command(cmd)
        r.changed = True


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import sys

from reemote.execute import execute
from reemote.utilities.parse_kwargs_string import parse_kwargs_string
from reemote.utilities.validate_root_class_name_and_get_root_class import validate_root_class_name_and_get_root_class
from reemote.utilities.verify_source_file_contains_valid_class import verify_source_file_contains_valid_class
from reemote.utilities.produce_grid import produce_grid
from reemote.utilities.produce_json import produce_json
from reemote.utilities.produce_output_grid import produce_output_grid


async def run_the_deploy(inventory, execution_report, stdout_report, sources):
    if sources.source != "/":

        if sources.source and sources.deployment:
            if not verify_source_file_contains_valid_class(sources.source, sources.deployment):
                sys.exit(1)

        # Verify the source and class
        if sources.source and sources.deployment:
            root_class = validate_root_class_name_and_get_root_class(sources.deployment, sources.source)

        try:
            # Parse parameters into kwargs
            kwargs = parse_kwargs_string(sources.kwargs)
            responses = []
            responses = await execute(inventory.inventory, root_class(**kwargs))

        except NameError:
            # Handle the case where the variable is not defined
            print("The variable 'root_class' is not defined.")
            sys.exit(1)


        columns, rows = produce_grid(produce_json(responses))
        execution_report.set(columns, rows)
        execution_report.execution_report.refresh()
        columns, rows = produce_output_grid(produce_json(responses))
        stdout_report.set(columns, rows)
        stdout_report.execution_report.refresh()



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command


class Assert:
    """
    A class to encapsulate the functionality of asserting expressions in Ansible-style fashion.
    It allows users to specify expressions that must evaluate to true, with optional custom messages
    for both success and failure cases. The module supports quiet mode to reduce output verbosity.

    Attributes:
        that (list): A list of string expressions to evaluate.
        fail_msg (str): Custom message to display on assertion failure.
        success_msg (str): Custom message to display on assertion success.
        quiet (bool): If True, suppresses verbose output.

    **Examples:**

    .. code:: python

        # Simple assertion
        yield Assert(that=["ansible_os_family != 'RedHat'"])

        # Multiple conditions with custom messages
        yield Assert(
            that=[
                "my_param <= 100",
                "my_param >= 0"
            ],
            fail_msg="'my_param' must be between 0 and 100",
            success_msg="'my_param' is between 0 and 100"
        )

        # Quiet mode to reduce output
        yield Assert(
            that=["my_param <= 100", "my_param >= 0"],
            quiet=True
        )

    Usage:
        This class is designed to be used in a generator-based workflow where assertions are yielded for evaluation.

    Notes:
        - Expressions follow the same syntax as Ansible's `when` statements.
        - Both `fail_msg` and `success_msg` support templating.
        - The `quiet` parameter can be used to minimize output in large-scale deployments.
    """

    def __init__(self,
                 that: list,
                 fail_msg: str = None,
                 success_msg: str = None,
                 quiet: bool = False):

        self.that = that
        self.fail_msg = fail_msg
        self.success_msg = success_msg
        self.quiet = quiet

    def __repr__(self):
        return (f"Assert(that={self.that!r}, "
                f"fail_msg={self.fail_msg!r}, "
                f"success_msg={self.success_msg!r}, "
                f"quiet={self.quiet!r})")

    def execute(self):
        # Convert the expressions into a format suitable for evaluation
        # This is a simplified implementation - in practice, this would need
        # to handle variable substitution and complex expression evaluation
        expressions = self.that if isinstance(self.that, list) else [self.that]

        # For demonstration purposes, we'll construct a shell command that evaluates the expressions
        # In a real implementation, this would interface with the Ansible evaluation engine
        expr_string = " && ".join(expressions)

        # Build command with appropriate flags
        cmd_parts = ["test"]

        for expr in expressions:
            # Convert common Ansible expressions to shell equivalents
            # This is a simplified mapping - real implementation would be more comprehensive
            shell_expr = expr.replace("==", "-eq").replace("!=", "-ne")
            cmd_parts.extend(["'", shell_expr, "'"])

        cmd = " ".join(cmd_parts)

        # Execute the command
        r = yield Command(
            cmd=cmd,
            guard=True  # Always execute assertions
        )

        # Set changed status (assertions typically don't change state)
        r.changed = False

        # Handle messaging based on result and quiet flag
        if not self.quiet:
            if r.cp.returncode == 0 and self.success_msg:
                print(self.success_msg)
            elif r.cp.returncode != 0 and (self.fail_msg or self.fail_msg != ""):
                fail_message = self.fail_msg or f"Assertion failed: {expr_string}"
                print(fail_message)
            elif r.cp.returncode != 0:
                print(f"Assertion failed: {expr_string}")



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from nicegui import native

from reemote.gui.classes.inventory_upload import Inventory_upload
from reemote.gui.classes.versions import Versions
from reemote.gui.classes.manger import Manager
from reemote.gui.classes.stdout_report import Stdout_report
from reemote.gui.classes.execution_report import Execution_report
from reemote.gui.classes.source_upload import Sources_upload
from reemote.gui.classes.file_path import File_path
from reemote.gui.classes.ad_hoc import Ad_Hoc

from reemote.gui.managers.command_manager import command_manager
from reemote.gui.managers.deployment_manager import deployment_manager
from reemote.gui.managers.file_manager import file_manager
from reemote.gui.managers.inventory_manager import inventory_manager
from reemote.gui.managers.package_manager import package_manager
from nicegui import ui

@ui.page('/')
def index():
    inventory=Inventory_upload()
    versions=Versions()
    manager=Manager()
    stdout_report=Stdout_report()
    execution_report=Execution_report()
    source=Sources_upload()
    file_path = File_path()
    ad_hoc = Ad_Hoc()

    with ui.header().classes(replace='row items-center') as header:
        with ui.tabs().classes('w-full') as tabs:
            ui.tab('Inventory Manager')
            ui.tab('Deployment Manager')
            ui.tab('Command Manager')
            ui.tab('File Manager')
            ui.tab('Package Manager')

    with ui.tab_panels(tabs,value='Inventory Manager').classes('w-full') as panels:
        with ui.tab_panel('Inventory Manager'):
            ui.label('Inventory Manager')
            inventory_manager(tabs, inventory, versions, manager, stdout_report, execution_report)
        with ui.tab_panel('Deployment Manager'):
            ui.label('Deployment Manager')
            deployment_manager(tabs, inventory, versions, manager, stdout_report, execution_report, source)
        with ui.tab_panel('Command Manager'):
            ui.label('Command Manager')
            command_manager(tabs, inventory, versions, manager, stdout_report, execution_report, ad_hoc)
        with ui.tab_panel('File Manager'):
            ui.label('File Manager')
            file_manager(tabs, inventory, versions, manager, stdout_report, execution_report, file_path)
        with ui.tab_panel('Package Manager'):
            ui.label('Package Manager')
            package_manager(tabs, inventory, versions, manager, stdout_report, execution_report)

    with ui.footer() as footer:
        stdout_report.execution_report()
        execution_report.execution_report()

ui.run(
    title="Reemotecontrol",
    reload=False,
    port=native.find_open_port(),
    storage_secret='private key to secure the browser session cookie'
)





# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import os
import tempfile
import hashlib
from pathlib import Path
import jinja2


from reemote.operations.sftp.read_file import Read_file
from reemote.operations.sftp.write_file import Write_file
from reemote.operations.sftp.setstat import Setstat


class Template():
    """
    A class to manage builtin templating on remote servers using SFTP operations.

    This class allows you to render Jinja2 templates and transfer them to remote servers.
    It supports template rendering with variables, backup creation, validation commands,
    and setting builtin attributes after transfer.

    Attributes:
        src (str): Path to the Jinja2 template builtin on the local system.
        dest (str): Path where the rendered template should be placed on the remote server.
        vars (dict, optional): Variables to use when rendering the template. Defaults to ``None``.
        backup (bool, optional): Whether to create a backup of the existing builtin. Defaults to ``False``.
        force (bool, optional): Whether to replace the builtin if content differs. Defaults to ``False``.
        lstrip_blocks (bool, optional): Strip leading spaces and tabs from blocks. Defaults to ``False``.
        trim_blocks (bool, optional): Remove first newline after a block. Defaults to ``False``.
        newline_sequence (str, optional): Newline sequence to use ('\\n', '\\r', '\\r\\n'). Defaults to ``None``.
        output_encoding (str, optional): Encoding for the output builtin. Defaults to ``None``.
        block_start_string (str, optional): String marking block start. Defaults to ``None``.
        block_end_string (str, optional): String marking block end. Defaults to ``None``.
        variable_start_string (str, optional): String marking variable start. Defaults to ``None``.
        variable_end_string (str, optional): String marking variable end. Defaults to ``None``.
        comment_start_string (str, optional): String marking comment start. Defaults to ``None``.
        comment_end_string (str, optional): String marking comment end. Defaults to ``None``.
        validate (str, optional): Validation command to run before final copy. Defaults to ``None``.
        attrs (dict, optional): File attributes to set after writing the builtin. Defaults to ``None``.

    **Examples:**

    .. code:: python

        # Create temporary build directory
        yield Tempfile(
            state="directory",
            suffix="build",
        )

        # Create temporary file
        yield Tempfile(
            state="file",
            suffix="temp",
        )
        # register: tempfile_1

        # Create a temporary file with a specific prefix
        yield Tempfile(
            state="file",
            suffix="txt",
            prefix="myfile_",
        )

        # Use the registered var and the file module to remove the temporary file
        yield File(
            path="{{ tempfile_1.path }}",
            state="absent",
        )
        # when: tempfile_1.path is defined

    Usage:
        This class is designed to be used in a generator-based workflow where
        commands are yielded for execution.
    """

    def __init__(self,
                 src="",
                 dest="",
                 vars=None,
                 backup=False,
                 force=True,
                 lstrip_blocks=False,
                 trim_blocks=True,
                 newline_sequence='\n',
                 output_encoding='utf-8',
                 block_start_string='{%',
                 block_end_string='%}',
                 variable_start_string='{{',
                 variable_end_string='}}',
                 comment_start_string='{#',
                 comment_end_string='#}',
                 validate=None,
                 attrs=None):

        self.src = src
        self.dest = dest
        self.vars = vars or {}
        self.backup = backup
        self.force = force
        self.lstrip_blocks = lstrip_blocks
        self.trim_blocks = trim_blocks
        self.newline_sequence = newline_sequence
        self.output_encoding = output_encoding
        self.block_start_string = block_start_string
        self.block_end_string = block_end_string
        self.variable_start_string = variable_start_string
        self.variable_end_string = variable_end_string
        self.comment_start_string = comment_start_string
        self.comment_end_string = comment_end_string
        self.validate = validate
        self.attrs = attrs

    def _calculate_checksum(self, content):
        """Calculate MD5 checksum of content."""
        if isinstance(content, dict):
            content = str(content)  # Convert dict to string for checksum
        return hashlib.md5(content.encode('utf-8')).hexdigest()

    def _read_template_file(self):
        """Read the template builtin with UTF-8 encoding."""
        try:
            with open(self.src, 'r', encoding='utf-8') as f:
                return f.read()
        except Exception as e:
            raise Exception(f"Failed to read template builtin '{self.src}': {str(e)}")

    def _render_template(self, template_content):
        """Render the Jinja2 template with the provided variables."""
        try:
            # Configure Jinja2 environment
            env = jinja2.Environment(
                loader=jinja2.BaseLoader(),
                trim_blocks=self.trim_blocks,
                lstrip_blocks=self.lstrip_blocks,
                block_start_string=self.block_start_string,
                block_end_string=self.block_end_string,
                variable_start_string=self.variable_start_string,
                variable_end_string=self.variable_end_string,
                comment_start_string=self.comment_start_string,
                comment_end_string=self.comment_end_string
            )

            template = env.from_string(template_content)
            rendered_content = template.render(**self.vars)

            # Handle newline sequence conversion
            if self.newline_sequence != '\n':
                rendered_content = rendered_content.replace('\n', self.newline_sequence)

            return rendered_content

        except jinja2.TemplateError as e:
            raise Exception(f"Template rendering failed: {str(e)}")
        except Exception as e:
            raise Exception(f"Unexpected error during template rendering: {str(e)}")

    def _ensure_string_content(self, content):
        """Ensure content is a string, converting if necessary."""
        if isinstance(content, dict):
            # Convert dictionary to JSON string
            import json
            return json.dumps(content, indent=2)
        elif isinstance(content, list):
            # Convert list to string representation
            return str(content)
        elif not isinstance(content, str):
            # Convert any other type to string
            return str(content)
        return content

    def execute(self):
        """
        :no-index:
        """
        try:
            # Step 1: Read and render the template
            template_content = self._read_template_file()
            rendered_content = self._render_template(template_content)

            # Ensure content is a string
            rendered_content = self._ensure_string_content(rendered_content)
            new_checksum = self._calculate_checksum(rendered_content)

            # Step 2: Check if remote builtin exists and compare content
            remote_exists = False
            remote_checksum = None
            remote_content = None

            try:
                r = yield Read_file(path=self.dest)
                remote_content = r.cp.stdout
                if remote_content:
                    remote_exists = True
                    if isinstance(remote_content, bytes):
                        remote_content = remote_content.decode('utf-8')
                    remote_content = self._ensure_string_content(remote_content)
                    remote_checksum = self._calculate_checksum(remote_content)
            except Exception:
                # File doesn't exist or can't be read
                remote_exists = False

            # Step 3: Check if changes are needed
            content_identical = remote_checksum == new_checksum

            if not self.force and remote_exists:
                print(f"File exists and force=False, skipping: {self.dest}")
                if self.attrs:
                    yield Setstat(path=self.dest, attrs=self.attrs)
                return

            if content_identical and remote_exists:
                print(f"Content unchanged, skipping: {self.dest}")
                if self.attrs:
                    yield Setstat(path=self.dest, attrs=self.attrs)
                return

            # Step 4: Create backup if requested and builtin exists
            if self.backup and remote_exists and remote_content:
                print(f"Creating backup at {self.dest}.backup")
                yield Write_file(path=f"{self.dest}.backup", text=remote_content)
                if self.attrs:
                    yield Setstat(path=f"{self.dest}.backup", attrs=self.attrs)

            # Step 5: Validate content if validation command provided
            if self.validate:
                print(f"Validating template with: {self.validate}")
                validation_passed = yield from self._validate_content(rendered_content)
                if not validation_passed:
                    raise Exception(f"Validation command failed: {self.validate}")

            # Step 6: Write the rendered content to remote builtin
            print(f"Writing template to: {self.dest}")
            yield Write_file(path=self.dest, text=rendered_content)

            # Step 7: Set builtin attributes if provided
            if self.attrs:
                print(f"Setting attributes for {self.dest}: {self.attrs}")
                yield Setstat(path=self.dest, attrs=self.attrs)

            print(f"Template successfully applied to: {self.dest}")

        except Exception as e:
            raise Exception(f"Template operation failed: {str(e)}")

    def _validate_content(self, content):
        """Validate the rendered content using the validation command."""
        if not self.validate:
            return True

        try:
            # Ensure content is string for validation
            content = self._ensure_string_content(content)

            # Create temporary builtin with rendered content
            with tempfile.NamedTemporaryFile(mode='w', encoding='utf-8', delete=False) as temp_file:
                temp_file.write(content)
                temp_path = temp_file.name

            # Replace %s in validate command with temp builtin path
            validate_cmd = self.validate.replace('%s', temp_path)

            # Execute validation command
            result = yield Command(command=validate_cmd)

            # Clean up temp builtin
            os.unlink(temp_path)

            return result.return_code == 0

        except Exception as e:
            raise Exception(f"Validation failed: {str(e)}")


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from typing import List
from reemote.command import Command

class Remove_apt_key:
    """
    Handles removing an apt GPG key from the system.

    This class is responsible for removing a GPG key from the system's
    apt keyring. It supports additional options such as using
    elevated privileges with `sudo` or `su`.

    Attributes:
        name (str): A descriptive name for the key being removed, used to
            identify the keyring builtin.
        guard (bool): Whether to add an execution guard for the
            operation. Defaults to True.
        sudo (bool): Indicates if the operation should be executed
            with `sudo` privileges. Defaults to True (typically required).
        su (bool): Indicates if the operation should be executed
            with `su` privileges. Defaults to False.
    """

    def __init__(self,
                 name: str = "",
                 guard: bool = True,
                 sudo: bool = True,
                 su: bool = False):
        self.name: str = name
        self.guard: bool = guard
        self.sudo: bool = sudo
        self.su: bool = su

    def __repr__(self) -> str:
        return (f"Remove_apt_key("
                f"name={self.name!r}, "
                f"guard={self.guard!r}, "                                
                f"sudo={self.sudo!r}, "
                f"su={self.su!r})")

    def execute(self):
        keyring_file = f"/usr/share/keyrings/{self.name.lower().replace(' ', '-')}.gpg"
        yield Command(
            f"rm -f {keyring_file}",
            guard=self.guard,
            sudo=self.sudo,
            su=self.su
        )



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command

class Get_sudo_password:
    """A reemote command to retrieve the sudo password from the global context.

    This class is designed to be used within a reemote execution plan. It
    provides a way to access the sudo password without hardcoding it
    or passing it directly as a parameter to other functions.

    When its `execute` method is called, it yields a local command. The
    reemote runner processes this command by invoking a callback function
    that retrieves the password from the `global_info` dictionary. The
    runner is responsible for populating `global_info['sudo_password']`
    before this command runs.
    """
    def __repr__(self):
        return f"Get_sudo_password()"

    @staticmethod
    async def _get_sudo_password_callback(host_info, global_info, command, cp, caller):
        return global_info["sudo_password"]

    def execute(self):
        r = yield Command(
            f"{self}",
            local=True,
            callback=self._get_sudo_password_callback,
            caller=self
        )



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import asyncssh
from reemote.command import Command
from typing import Union, List, Optional
import tempfile
import os


class Copy:
    """
    A class to handle secure builtin copying between remote hosts using SCP (Secure Copy Protocol).

    This class provides functionality to copy builtin or directories between remote hosts,
    from source hosts to destination hosts. It supports various SCP options including
    preserving builtin attributes and recursive directory copying.

    Attributes:
        srcpaths (Union[str, List[str]]): Source builtin or directory path(s). Can be a single
            string path or a list of paths. Supports host:path format for remote sources.
        dstpath (str): Destination path where builtin will be copied. Supports host:path format
            for remote destinations.
        src_hosts (List[str], optional): List of source host identifiers. If None or empty,
            operation will attempt to use all available hosts as sources.
        dst_hosts (List[str], optional): List of destination host identifiers. If None or empty,
            operation will attempt to use all available hosts as destinations.
        preserve (bool): If True, preserves builtin modification times, access times,
            and modes from the original builtin. Defaults to False.
        recurse (bool): If True, recursively copies entire directories. Defaults to False.
        block_size (int): Block size used for builtin transfers in bytes. Defaults to 16384.
        port (int): SSH port to use for connections. Defaults to 22.

    **Examples:**

    .. code:: python

        # Copy builtin from one host to another
        yield Copy(
            srcpaths='/home/user/*.txt',
            dstpath='/home/user/',
            src_hosts=["10.156.135.16"],
            dst_hosts=["10.156.135.17"],
            recurse=True
        )
        # Copy multiple builtin between hosts
        yield Copy(
            srcpaths=['/var/log/app.log', '/tmp/debug.log'],
            dstpath='backup-server:/backup/logs/',
            src_hosts=["host1", "host2"],
            dst_hosts=["backup-server"]
        )

    Note:
        - The copy operation requires proper SSH credentials and permissions on both source and destination hosts
        - Wildcard patterns are supported in source paths
        - For host-to-host copies, the operation will be executed from the destination host
    """

    def __init__(self,
                 srcpaths: Union[str, List[str]],
                 dstpath: str,
                 src_hosts: List[str] = None,
                 dst_hosts: List[str] = None,
                 preserve: bool = False,
                 recurse: bool = False,
                 block_size: int = 16384,
                 port: int = 22):
        self.srcpaths = srcpaths
        self.dstpath = dstpath
        self.src_hosts = src_hosts
        self.dst_hosts = dst_hosts
        self.preserve = preserve
        self.recurse = recurse
        self.block_size = block_size
        self.port = port

    def __repr__(self):
        return f"Copy(srcpaths={self.srcpaths!r}, dstpath={self.dstpath!r}, src_hosts={self.src_hosts!r}, dst_hosts={self.dst_hosts!r}, preserve={self.preserve}, recurse={self.recurse})"

    @staticmethod
    async def _copy_callback(host_info, global_info, command, cp, caller):
        """Static callback method for builtin copy between hosts"""

        # Validate host_info (matching Download error handling)
        required_keys = ['host', 'username', 'password']
        for key in required_keys:
            if key not in host_info or host_info[key] is None:
                raise ValueError(f"Missing or invalid value for '{key}' in host_info.")

        # Validate caller attributes (matching Download error handling)
        if caller.srcpaths is None:
            raise ValueError("The 'srcpaths' attribute of the caller cannot be None.")
        if caller.dstpath is None:
            raise ValueError("The 'dstpath' attribute of the caller cannot be None.")

        # This operation should run on destination hosts to pull builtin from source hosts
        is_dest_host = (caller.dst_hosts is None or
                        not caller.dst_hosts or
                        host_info["host"] in caller.dst_hosts)

        if not is_dest_host:
            return  # Only run on destination hosts

        try:
            # Create connection parameters for destination host
            connect_kwargs = {
                'host': host_info['host'],
                'username': host_info.get('username'),
                'password': host_info.get('password'),
                'client_keys': host_info.get('client_keys'),
                'known_hosts': host_info.get('known_hosts')
            }

            # Remove None values
            connect_kwargs = {k: v for k, v in connect_kwargs.items() if v is not None}

            # Set port if specified and different from default
            if caller.port != 22:
                connect_kwargs['port'] = caller.port

            print(f"Connecting to destination host {host_info['host']} for copy operation")

            async with asyncssh.connect(**connect_kwargs) as conn:
                print(f"Connected successfully to destination host {host_info['host']}")

                # Prepare source paths with proper host prefixes
                if isinstance(caller.srcpaths, list):
                    srcpaths = []
                    for srcpath in caller.srcpaths:
                        # Add source host prefix if not already present
                        if ':' not in srcpath and caller.src_hosts:
                            # Use the first source host if multiple specified
                            source_host = caller.src_hosts[0]
                            srcpaths.append(f"{source_host}:{srcpath}")
                        else:
                            srcpaths.append(srcpath)
                else:
                    if ':' not in caller.srcpaths and caller.src_hosts:
                        source_host = caller.src_hosts[0]
                        srcpaths = f"{source_host}:{caller.srcpaths}"
                    else:
                        srcpaths = caller.srcpaths

                # Prepare destination path (local to destination host)
                if ':' in caller.dstpath:
                    # Extract path if destination has host prefix
                    dstpath = caller.dstpath.split(':', 1)[1]
                else:
                    dstpath = caller.dstpath

                print(f"Copying from source: {srcpaths} -> destination: {dstpath} on {host_info['host']}")

                # Perform the SCP copy operation (pulling from source to destination)
                await asyncssh.scp(
                    srcpaths,
                    dstpath,
                    preserve=caller.preserve,
                    recurse=caller.recurse,
                    block_size=caller.block_size
                )

                print(f"Copy completed successfully to destination host {host_info['host']}")

        except (OSError, asyncssh.Error) as exc:
            print(f'SCP copy failed on destination host {host_info["host"]}: {str(exc)}')
            # Provide more detailed error information
            if "Permission denied" in str(exc):
                print(f"Check that user {host_info.get('username', 'current user')} has:")
                print(f" - Read permission on source paths: {caller.srcpaths}")
                print(f" - Write permission on destination path: {caller.dstpath}")
                print(f" - Proper SSH key authentication configured")
            raise

    def execute(self):
        """
        Execute the copy operation between hosts.

        Yields:
            Operation: An Operation object that handles the asynchronous copy process.

        Returns:
            Command: The operation object with executed and changed flags set to True.
        """
        r = yield Command(f"{self}", local=True, callback=self._copy_callback, caller=self)
        r.executed = True
        r.changed = True
        return r



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import asyncssh
import asyncio
import sys
import posixpath
from reemote.command import Command
from typing import cast


class Rmtree:
    """
    A class to encapsulate the functionality of rmtree (recursively remove directory tree).

    Attributes:
        path (str): The directory path to remove recursively.
        ignore_errors (bool): Whether to ignore errors during removal.

    **Examples:**

    .. code:: python

        yield Rmtree(path='/home/user/hfs', ignore_errors=False)

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.
    """

    def __init__(self,
                 path: str,
                 ignore_errors: bool = False):
        self.path = path
        self.ignore_errors = ignore_errors

    def __repr__(self):
        return (f"Rmtree(path={self.path!r}, "
                f"ignore_errors={self.ignore_errors!r})")

    @staticmethod
    async def _rmtree_callback(host_info, global_info, command, cp, caller):
        """Static callback method for recursive directory removal"""

        # Validate host_info (matching Rmdir error handling)
        required_keys = ['host', 'username', 'password']
        for key in required_keys:
            if key not in host_info or host_info[key] is None:
                raise ValueError(f"Missing or invalid value for '{key}' in host_info.")

        # Validate caller attributes (matching Rmdir error handling)
        if caller.path is None:
            raise ValueError("The 'path' attribute of the caller cannot be None.")

        try:
            # Connect to the SSH server and recursively remove the directory
            async with asyncssh.connect(**host_info) as conn:
                async with conn.start_sftp_client() as sftp:
                    # Use the rmtree method provided by asyncssh
                    await sftp.rmtree(path=caller.path,
                                      ignore_errors=caller.ignore_errors)
        except (OSError, asyncssh.Error) as exc:
            raise  # Re-raise the exception to handle it in the caller

    def execute(self):
        r = yield Command(f"{self}", local=True, callback=self._rmtree_callback, caller=self)
        r.executed = True
        r.changed = False
        return r



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import asyncssh
from asyncssh import SFTPAttrs
from reemote.command import Command


class Makedirs:
    """
    A class to encapsulate the functionality of makedirs (recursive directory creation).

    Attributes:
        path (str): The directory path to create recursively.
        attrs (SFTPAttrs): asyncssh SFTPAttrs object for directory attributes.
        exist_ok (bool): Whether to raise an error if the target directory already exists.

    **Examples:**

    .. code:: python

        yield Makedirs(path='/home/user/hfs/subdir1/subdir2',
             attrs=SFTPAttrs(permissions=0o755),
             exist_ok=True,
        )

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        This will create all intermediate directories in the path if they don't exist.
    """

    @staticmethod
    async def _makedirs_callback(host_info, global_info, command, cp, caller):
        """Static callback method for directory creation"""

        # Validate host_info
        required_keys = ['host', 'username', 'password']
        for key in required_keys:
            if key not in host_info or host_info[key] is None:
                raise ValueError(f"Missing or invalid value for '{key}' in host_info.")

        # Validate caller attributes
        if caller.path is None:
            raise ValueError("The 'path' attribute of the caller cannot be None.")

        try:
            async with asyncssh.connect(**host_info) as conn:
                async with conn.start_sftp_client() as sftp:
                    # Create the remote directory recursively
                    await sftp.makedirs(path=caller.path, attrs=caller.attrs, exist_ok=caller.exist_ok)
                    return f"Successfully created directory {caller.path} on {host_info['host']}"
        except (OSError, asyncssh.Error) as exc:
            raise  # Re-raise the exception to handle it in the caller

    def __init__(self,
                 path: str,
                 attrs: SFTPAttrs = None,
                 exist_ok: bool = False):
        self.path = path

        # Set default SFTPAttrs if none provided
        self.attrs = attrs if attrs is not None else SFTPAttrs()
        self.exist_ok = exist_ok

    def __repr__(self):
        return (f"Makedirs(path={self.path!r}, "
                f"attrs={self.attrs!r}, "
                f"exist_ok={self.exist_ok!r})")

    def execute(self):
        r = yield Command(f"{self}", local=True, callback=self._makedirs_callback, caller=self)
        r.executed = True
        r.changed = False
        return r



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from typing import List


def validate_list_of_strings(x) -> List[str]:
    """Verifies that the input is a list containing only string elements.

    This function acts as a type-guard, ensuring that an input value is
    a list composed exclusively of strings. It is primarily used to
    validate data before it is processed by other parts of an application
    that rely on this specific data structure.

    The function performs two key checks:
    - It verifies that the input `x` is a `list`.
    - It iterates through the list to confirm that every element is a `str`.

    If either check fails, a `TypeError` is raised with a message
    indicating the nature of the validation failure.

    Args:
        x (Any): The input value to be validated.

    Returns:
        List[str]: The original list, if it passes validation.

    Raises:
        TypeError: If the input is not a list, or if the list contains
            any non-string elements.

    Examples:
        >>> validate_list_of_strings(["pkg1", "pkg2"])
        ['pkg1', 'pkg2']
        >>> validate_list_of_strings("package")
        Traceback (most recent call last):
        ...
        TypeError: Input must be a list of strings
        >>> validate_list_of_strings(["item", 123])
        Traceback (most recent call last):
        ...
        TypeError: All elements in the list must be strings
    """
    if not isinstance(x, list):
        raise TypeError("Input must be a list of strings")
    
    # Validate that all elements are strings
    if not all(isinstance(item, str) for item in x):
        raise TypeError("All elements in the list must be strings")
    
    return x


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command

class Get_cp:
    """A command-like object to retrieve the current command processor.

    This class is designed to be used within the reemote command execution
    framework. When its `execute` method is called, it yields a special
    local command. The framework intercepts this command and, via a
    callback, provides the instance of the currently running command
    processor (`cp`).

    This is useful for other commands that need to introspect or interact
    with the command processor that is executing them.
    """
    def __repr__(self):
        return f"Get_cp()"

    async def _get_cp_callback(self, host_info, global_info, command, cp, caller):
        return cp

    def execute(self):
        r = yield Command(
            f"{self}",
            local=True,
            callback=self._get_cp_callback,
            caller=self
        )


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#

class Install_rust:
    def execute(self):
        from reemote.facts.server.get_os import Get_OS
        os = (yield Get_OS("NAME")).cp.stdout
        print(os)
        if "Alpine" in os:
            from reemote.operations.apk.packages import Operation_packages
            from reemote.operations.apk.update import Update
            r = yield Update()
            from reemote.operations.server.shell import Shell
            r = yield Operation_packages(packages=["bash", "curl", "gcc", "musl-dev", "openssl-dev"], present=True, su=True)
        if "Ubuntu" in os:
            from reemote.operations.apk.packages import Operation_packages
            from reemote.operations.apk.update import Update
            r = yield Update()
            yield Operation_packages(packages=["build-essential", "curl"], present=True, su=True)
        r = yield Shell("curl --proto '=https' --tlsv1.2 -sSf https://sh.rustup.rs | sh -s -- -y")



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command


class Fetch:
    """
    A class to encapsulate the functionality of fetching files from remote nodes.
    This module works like copy, but in reverse. It is used for fetching files
    from remote machines and storing them locally in a file tree, organized by hostname.

    Files that already exist at dest will be overwritten if they are different than the src.
    This module is also supported for Windows targets.

    Attributes:
        src (str): The file on the remote system to fetch. This must be a file, not a directory.
        dest (str): A directory to save the file into.
        fail_on_missing (bool): When set to true, the task will fail if the remote file cannot be read.
        flat (bool): Allows you to override the default behavior of appending hostname/path/to/file to the destination.
        validate_checksum (bool): Verify that the source and destination checksums match after the files are fetched.
        guard (bool): If `False` the commands will not be executed.

    **Examples:**

    .. code:: python

        # Store file into /tmp/fetched/host.example.com/tmp/somefile
        r = yield Fetch(src="/tmp/somefile", dest="/tmp/fetched")

        # Specifying a path directly
        r = yield Fetch(src="/tmp/somefile", dest="/tmp/prefix-{}".format(inventory_hostname), flat=True)

        # Specifying a destination path
        r = yield Fetch(src="/tmp/uniquefile", dest="/tmp/special/", flat=True)

        # Storing in a path relative to the playbook
        r = yield Fetch(src="/tmp/uniquefile", dest="special/prefix-{}".format(inventory_hostname), flat=True)

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        - When running fetch with become, the slurp module will also be used to fetch the contents
          of the file for determining the remote checksum. This effectively doubles the transfer size.
        - Files are organized by hostname in the destination directory unless flat=True is specified.
    """

    def __init__(self,
                 src: str,
                 dest: str,
                 fail_on_missing: bool = True,
                 flat: bool = False,
                 validate_checksum: bool = True,
                 guard: bool = True):

        self.src = src
        self.dest = dest
        self.fail_on_missing = fail_on_missing
        self.flat = flat
        self.validate_checksum = validate_checksum
        self.guard = guard

    def __repr__(self):
        return (f"Fetch(src={self.src!r}, "
                f"dest={self.dest!r}, "
                f"fail_on_missing={self.fail_on_missing!r}, "
                f"flat={self.flat!r}, "
                f"validate_checksum={self.validate_checksum!r}, "
                f"guard={self.guard!r})")

    def execute(self):
        # Construct the ansible fetch command
        cmd_parts = ["ansible.builtin.fetch"]
        cmd_parts.append(f"src={self.src}")
        cmd_parts.append(f"dest={self.dest}")

        if not self.fail_on_missing:
            cmd_parts.append("fail_on_missing=false")

        if self.flat:
            cmd_parts.append("flat=true")

        if not self.validate_checksum:
            cmd_parts.append("validate_checksum=false")

        cmd = " ".join(cmd_parts)
        r = yield Command(cmd, guard=self.guard)
        r.changed = True


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command


class DpkgSelections:
    """
    A class to manage dpkg package selection states using `dpkg --get-selections` and `dpkg --set-selections`.

    This class allows setting the selection state of a Debian package without installing/removing it.
    It supports standard dpkg selection states such as 'install', 'hold', 'deinstall', and 'purge'.

    Attributes:
        name (str): The name of the package.
        selection (str): The desired selection state for the package.
                         Choices: 'install', 'hold', 'deinstall', 'purge'.
        guard (bool): If `False`, the command will not be executed.

    Examples:
        .. code:: python

            # Hold a package to prevent upgrades
            r = yield DpkgSelections(name="python", selection="hold")
            print(r.cp.stdout)

            # Allow a package to be upgraded again
            r = yield DpkgSelections(name="python", selection="install")
            print(r.cp.stdout)

    Notes:
        - This module only changes the selection state; it does not install or remove packages.
          Use the apt module for actual package management.
        - Only works on Debian-based systems where dpkg is available.
    """

    VALID_SELECTIONS = {"install", "hold", "deinstall", "purge"}

    def __init__(self,
                 name: str,
                 selection: str,
                 guard: bool = True):
        if selection not in self.VALID_SELECTIONS:
            raise ValueError(
                f"Invalid selection '{selection}'. Must be one of {self.VALID_SELECTIONS}."
            )

        self.name = name
        self.selection = selection
        self.guard = guard

    def __repr__(self):
        return (
            f"DpkgSelections(name={self.name!r}, "
            f"selection={self.selection!r}, "
            f"guard={self.guard!r})"
        )

    def execute(self):
        # Construct the command to set the dpkg selection
        cmd_str = f"echo '{self.name} {self.selection}' | dpkg --set-selections"
        r = yield Command(cmd_str, guard=self.guard)
        r.changed = True


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command


class Apt_key:
    """
    A class to encapsulate the functionality of managing apt keys in Debian-based systems.
    It allows users to add or remove apt keys from keyring, optionally downloading them from
    a keyserver or URL, or using local keyfiles.

    Attributes:
        data (str): The keyfile contents to add to the keyring.
        file (str): The path to a keyfile on the remote server to add to the keyring.
        id (str): The identifier of the key.
        keyring (str): The full path to specific keyring file in /etc/apt/trusted.gpg.d/.
        keyserver (str): The keyserver to retrieve key from.
        state (str): Ensures that the key is present (added) or absent (revoked).
        url (str): The URL to retrieve key from.
        validate_certs (bool): If False, SSL certificates for the target url will not be validated.

    **Examples:**

    .. code:: python

        # Add an apt key by id from a keyserver
        r = yield AptKey(keyserver="keyserver.ubuntu.com",
                         id="36A1D7869245C8950F966E92D8576A8BA88D21E9")

        # Add an Apt signing key from URL
        r = yield AptKey(url="https://ftp-master.debian.org/keys/archive-key-6.0.asc",
                         state="present")

        # Remove an Apt specific signing key
        r = yield AptKey(id="0x9FED2BCBDCD29CDF762678CBAED4B06F473041FA",
                         state="absent")

        # Add a key from data content
        key_data = "-----BEGIN PGP PUBLIC KEY BLOCK-----\\n..."
        r = yield AptKey(data=key_data, state="present")

        # Add an Apt signing key to a specific keyring file
        r = yield AptKey(id="9FED2BCBDCD29CDF762678CBAED4B06F473041FA",
                         url="https://ftp-master.debian.org/keys/archive-key-6.0.asc",
                         keyring="/etc/apt/trusted.gpg.d/debian.gpg")

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        - The apt-key command used by this implementation has been deprecated.
        - Adding a new key requires an apt cache update.
        - Use full fingerprint (40 characters) key ids to avoid key collisions.
    """

    def __init__(self,
                 data: str = None,
                 file: str = None,
                 id: str = None,
                 keyring: str = None,
                 keyserver: str = None,
                 state: str = "present",
                 url: str = None,
                 validate_certs: bool = True,
                 guard: bool = True,
                 sudo: bool = True):  # apt-key typically requires sudo

        self.data = data
        self.file = file
        self.id = id
        self.keyring = keyring
        self.keyserver = keyserver
        self.state = state
        self.url = url
        self.validate_certs = validate_certs
        self.guard = guard
        self.sudo = sudo

    def __repr__(self):
        return (f"AptKey(data={self.data!r}, "
                f"file={self.file!r}, "
                f"id={self.id!r}, "
                f"keyring={self.keyring!r}, "
                f"keyserver={self.keyserver!r}, "
                f"state={self.state!r}, "
                f"url={self.url!r}, "
                f"validate_certs={self.validate_certs!r}, "
                f"guard={self.guard!r}, "
                f"sudo={self.sudo!r})")

    def execute(self):
        # Build the apt-key command
        cmd_parts = ["apt-key"]

        if self.keyring:
            cmd_parts.extend(["--keyring", self.keyring])

        if self.state == "absent":
            if not self.id:
                raise ValueError("id parameter is required when state is absent")
            cmd_parts.extend(["del", self.id])
        else:  # state == "present"
            if self.data:
                # For data, we need to echo it and pipe to apt-key
                cmd_parts = ["echo", repr(self.data), "|"] + cmd_parts + ["add", "-"]
            elif self.file:
                cmd_parts.extend(["add", self.file])
            elif self.keyserver and self.id:
                cmd_parts.extend(["adv", "--keyserver", self.keyserver, "--recv-keys", self.id])
            elif self.url:
                # Download and add from URL
                cmd_parts.extend(["adv", "--fetch-keys", self.url])
            elif self.id and not self.keyserver:
                # If only ID is provided, try to receive from default keyserver
                cmd_parts.extend(["adv", "--recv-keys", self.id])
            else:
                raise ValueError("Insufficient parameters to add key. Provide data, file, keyserver+id, or url.")

        cmd = " ".join(cmd_parts)
        r = yield Command(cmd, guard=self.guard, sudo=self.sudo)
        r.changed = True


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.utilities.produce_grid import produce_grid
from reemote.utilities.produce_json import produce_json
from reemote.utilities.produce_output_grid import produce_output_grid


async def upload_inventory(inventory, execution_report, stdout_report):
    # Start with the fixed column definition for "Command"
    columns = [{'headerName': 'Command', 'field': 'command'}]
    rows = []

    # Dynamically generate column definitions for each host
    for index, (host_info, _) in enumerate(inventory.inventory):
        host_ip = host_info['host']
        columns.append({'headerName': f'{host_ip} Executed', 'field': f'{host_ip.replace(".","_")}_executed'})
        columns.append({'headerName': f'{host_ip} Changed', 'field': f'{host_ip.replace(".","_")}_changed'})

    # columns, rows = produce_grid(produce_json(responses))
    execution_report.set(columns, rows)
    execution_report.execution_report.refresh()
    # columns, rows = produce_output_grid(produce_json(responses))
    stdout_report.set(columns, rows)
    stdout_report.execution_report.refresh()



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from typing import List
from reemote.operation_update import Operation_update
from reemote.commands.yum.upgrade import Upgrade
from reemote.facts.yum.get_packages import Get_packages


class Update(Operation_update):
    """
    A class to manage package update operations on a remote system using `yum`.

    This class provides functionality to update package lists on RHEL/CentOS systems
    using the yum package manager. It allows configuration of execution privileges and
    safety guards for package update operations.

    Attributes:
        guard (bool): If `False` the commands will not be executed.
        sudo (bool): If `True`, the commands will be executed with [sudo](file:///home/kim/reemote/reemote/command.py#L11-L11) privileges.
        su (bool): If `True`, the commands will be executed with [su](file:///home/kim/reemote/reemote/command.py#L12-L12) privileges.

    **Examples:**

    .. code:: python

        # Update package lists on all hosts
        r = yield Update()
        # Check if the operation was successful
        if r.cp.return_code == 0:
            print("Package lists updated successfully")

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        - Uses `yum update` command internally for updating package lists.
        - Inherits from [Operation_update](file:///home/kim/reemote/reemote/operation_update.py#L0-L63) base class.
        - This operation only updates the package lists but does not upgrade installed packages.
    """


    def __init__(self,
                 guard: bool = True,
                 sudo: bool = False,
                 su: bool = False):
        super().__init__(guard, sudo, su)

    def get_packages(self):
        return Get_packages()

    def update_packages(self, guard=None,sudo=None,su=None):
        from reemote.commands.yum.update import Update
        return Update(self.guard, self.sudo, self.su)



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from tabulate import tabulate

def _generate_table(data):
    """Transforms raw event data into a structured grid format.

    This internal helper function processes a list of event dictionaries,
    groups them by command, and pivots the data to create a table structure.
    For each unique host found in the data, it generates 'Executed' and
    'Changed' columns.

    The primary purpose is to group consecutive entries that share the same
    command string into a single row, showing the execution status across
    all relevant hosts.

    Args:
        data (list[dict]): A list of dictionaries, where each dictionary
            represents a single event. Each dictionary must contain keys:
            'host' (str), 'executed' (any), 'changed' (any), and 'op' (dict).
            The 'op' dictionary must contain a 'command' (str) key.

    Returns:
        tuple: A tuple containing two elements:
            - result (dict): A dictionary with 'columnDefs' and 'rowData',
              suitable for a grid component.
            - hosts (list[str]): A sorted list of unique hostnames
              found in the data.

    Raises:
        TypeError: If the input 'data' is not a list of dictionaries,
                   preventing host extraction.
    """
    # Step 1: Extract unique hosts
    try:
        hosts = sorted(set(entry['host'] for entry in data))
    except TypeError as e:
        print("Error: The 'data' variable must be a list of dictionaries.")
        raise e

    # Step 2: Initialize columnDefs and rowData
    result = {
        "columnDefs": [],
        "rowData": []
    }

    # Add the 'Command' column to columnDefs
    result["columnDefs"].append({"headerName": "Command", "field": "command"})

    # Add two columns for each host: Executed and Changed
    for host in hosts:
        result["columnDefs"].append({"headerName": f"{host} Executed", "field": f"{host.replace(".","_")}_executed"})
        result["columnDefs"].append({"headerName": f"{host} Changed", "field": f"{host.replace(".","_")}_changed"})

    # Step 3: Process data by grouping consecutive entries with the same command
    i = 0
    while i < len(data):
        # Get the current command
        command = data[i]['op']['command']

        # Create a new row for this command
        row = {"command": command}

        # Initialize all host columns as empty
        for h in hosts:
            row[f"{h}_executed"] = ""
            row[f"{h}_changed"] = ""

        # Process all consecutive entries with this same command
        while i < len(data) and data[i]['op']['command'] == command:
            entry = data[i]
            host = entry['host']
            executed = entry['executed']
            changed = entry['changed']

            # Add the current host's data
            row[f"{host.replace(".","_")}_executed"] = executed
            row[f"{host.replace(".","_")}_changed"] = changed

            i += 1  # Move to next entry

        # Add the completed row to rowData
        result["rowData"].append(row)

    return result, hosts

def generate_table(data):
    """Generates a human-readable ASCII grid table from event data.

     This function takes a list of event data, processes it into a tabular
     format, and then uses the `tabulate` library to create a formatted
     string representation of the table. It is ideal for displaying
     command execution results in a console or log builtin.

     It internally uses `_generate_table` to structure the data before
     formatting.

     Args:
         data (list[dict]): A list of event dictionaries. See the
             docstring for `_generate_table` for the required structure.

     Returns:
         str: A string containing the formatted ASCII table in 'grid' format.
     """
    result , hosts = _generate_table(data)

    # Step 4: Convert to tabulate format
    headers = [col['headerName'] for col in result["columnDefs"]]
    table_data = []
    for row in result["rowData"]:
        formatted_row = [row['command']]  # Start with command
        for host in hosts:
            executed = row[f"{host.replace(".","_")}_executed"]
            changed = row[f"{host.replace(".","_")}_changed"]
            formatted_row.append(executed if executed != "" else "")
            formatted_row.append(changed if changed != "" else "")
        table_data.append(formatted_row)

    # Step 5: Return the table
    return tabulate(table_data, headers=headers, tablefmt="grid")

def generate_grid(data):
    """Generates a human-readable ASCII grid table from event data.

    This function takes a list of event data, processes it into a tabular
    format, and then uses the `tabulate` library to create a formatted
    string representation of the table. It is ideal for displaying
    command execution results in a console or log builtin.

    It internally uses `_generate_table` to structure the data before
    formatting.

    Args:
        data (list[dict]): A list of event dictionaries. See the
            docstring for `_generate_table` for the required structure.

    Returns:
        str: A string containing the formatted ASCII table in 'grid' format.
    """
    result , hosts = _generate_table(data)
    return result["columnDefs"], result["rowData"],



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from typing import List
from reemote.command import Command

class Add_ppa:
    """
    Handles adding a Personal Package Archive (PPA) to the system.

    This class is responsible for configuring a PPA repository on
    a system. It supports additional options such as using
    elevated privileges with `sudo` or `su`. It also allows for a
    safety guard to control the execution of the repository addition.

    Attributes:
        ppa (str): The PPA repository string to be added.
        guard (bool): Whether to add an execution guard for the
            operation. Defaults to True.
        sudo (bool): Indicates if the operation should be executed
            with `sudo` privileges. Defaults to False.
        su (bool): Indicates if the operation should be executed
            with `su` privileges. Defaults to False.
    """

    def __init__(self,
                 ppa: str = "",
                 guard: bool = True,
                 sudo: bool = False,
                 su: bool = False):
        self.ppa: str = ppa
        self.guard: bool = guard
        self.sudo: bool = sudo
        self.su: bool = su

    def __repr__(self) -> str:
        return (f"Add_ppa("
                f"ppa={self.ppa!r}, "
                f"guard={self.guard!r}, "                                
                f"sudo={self.sudo!r}, "
                f"su={self.su!r})")

    def execute(self):
        yield Command(f"apt-add-repository -y {self.ppa}", guard=self.guard, sudo=self.sudo, su=self.su)




# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command


class User:
    """
    A class to encapsulate the functionality of managing user accounts in Unix-like operating systems.
    It allows users to create, modify, and remove user accounts with various attributes and options.

    Attributes:
        name (str): Name of the user to create, remove or modify.
        state (str): Whether the account should exist or not ('present' or 'absent').
        uid (int): Optionally sets the UID of the user.
        group (str): Optionally sets the user's primary group.
        groups (list): A list of supplementary groups which the user is also a member of.
        append (bool): If true, add the user to the groups specified in groups.
        comment (str): Optionally sets the description (aka GECOS) of user account.
        home (str): Optionally set the user's home directory.
        shell (str): Optionally set the user's shell.
        password (str): If provided, set the user's password.
        create_home (bool): Whether to create a home directory for the user.
        move_home (bool): If set to true when used with home, attempt to move the user's old home directory.
        system (bool): When creating an account, setting this to true makes the user a system account.
        force (bool): Forces removal of the user and associated directories.
        remove (bool): Attempts to remove directories associated with the user.
        generate_ssh_key (bool): Whether to generate a SSH key for the user.
        ssh_key_bits (int): Optionally specify number of bits in SSH key to create.
        ssh_key_file (str): Optionally specify the SSH key filename.
        ssh_key_comment (str): Optionally define the comment for the SSH key.
        ssh_key_passphrase (str): Set a passphrase for the SSH key.
        ssh_key_type (str): Optionally specify the type of SSH key to generate.
        update_password (str): Controls when to update passwords ('always' or 'on_create').
        expires (float): An expiry time for the user in epoch.
        password_lock (bool): Lock the password.
        password_expire_max (int): Maximum number of days between password change.
        password_expire_min (int): Minimum number of days between password change.
        password_expire_warn (int): Number of days of warning before password expires.
        password_expire_account_disable (int): Number of days after a password expires until the account is disabled.
        local (bool): Forces the use of "local" command alternatives.
        skeleton (str): Optionally set a home skeleton directory.
        umask (str): Sets the umask of the user.
        uid_min (int): Sets the UID_MIN value for user creation.
        uid_max (int): Sets the UID_MAX value for user creation.
        seuser (str): Optionally sets the seuser type on SELinux enabled systems.
        login_class (str): Optionally sets the user's login class.
        authorization (str): Sets the authorization of the user.
        role (str): Sets the role of the user.
        profile (str): Sets the profile of the user.
        hidden (bool): Optionally hide the user from the login window and system preferences (macOS only).

    **Examples:**

    .. code:: python

        # Add the user 'johnd' with a specific uid and a primary group of 'admin'
        r = yield User(name="johnd", comment="John Doe", uid=1040, group="admin")

        # Create a user 'johnd' with a home directory
        r = yield User(name="johnd", create_home=True)

        # Add the user 'james' with a bash shell, appending the group 'admins' and 'developers'
        r = yield User(name="james", shell="/bin/bash", groups=["admins", "developers"], append=True)

        # Remove the user 'johnd'
        r = yield User(name="johnd", state="absent", remove=True)

        # Create a 2048-bit SSH key for user jsmith
        r = yield User(name="jsmith", generate_ssh_key=True, ssh_key_bits=2048, ssh_key_file=".ssh/id_rsa")

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        - Commands are constructed based on the provided parameters and the target operating system.
        - The implementation uses underlying system tools like useradd, usermod, userdel, pw, dscl depending on the platform.
    """

    def __init__(self,
                 name: str,
                 state: str = "present",
                 uid: int = None,
                 group: str = None,
                 groups: list = None,
                 append: bool = False,
                 comment: str = None,
                 home: str = None,
                 shell: str = None,
                 password: str = None,
                 create_home: bool = True,
                 move_home: bool = False,
                 system: bool = False,
                 force: bool = False,
                 remove: bool = False,
                 generate_ssh_key: bool = False,
                 ssh_key_bits: int = None,
                 ssh_key_file: str = None,
                 ssh_key_comment: str = None,
                 ssh_key_passphrase: str = None,
                 ssh_key_type: str = "rsa",
                 update_password: str = "always",
                 expires: float = None,
                 password_lock: bool = None,
                 password_expire_max: int = None,
                 password_expire_min: int = None,
                 password_expire_warn: int = None,
                 password_expire_account_disable: int = None,
                 local: bool = False,
                 skeleton: str = None,
                 umask: str = None,
                 uid_min: int = None,
                 uid_max: int = None,
                 seuser: str = None,
                 login_class: str = None,
                 authorization: str = None,
                 role: str = None,
                 profile: str = None,
                 hidden: bool = None):

        self.name = name
        self.state = state
        self.uid = uid
        self.group = group
        self.groups = groups or []
        self.append = append
        self.comment = comment
        self.home = home
        self.shell = shell
        self.password = password
        self.create_home = create_home
        self.move_home = move_home
        self.system = system
        self.force = force
        self.remove = remove
        self.generate_ssh_key = generate_ssh_key
        self.ssh_key_bits = ssh_key_bits
        self.ssh_key_file = ssh_key_file
        self.ssh_key_comment = ssh_key_comment
        self.ssh_key_passphrase = ssh_key_passphrase
        self.ssh_key_type = ssh_key_type
        self.update_password = update_password
        self.expires = expires
        self.password_lock = password_lock
        self.password_expire_max = password_expire_max
        self.password_expire_min = password_expire_min
        self.password_expire_warn = password_expire_warn
        self.password_expire_account_disable = password_expire_account_disable
        self.local = local
        self.skeleton = skeleton
        self.umask = umask
        self.uid_min = uid_min
        self.uid_max = uid_max
        self.seuser = seuser
        self.login_class = login_class
        self.authorization = authorization
        self.role = role
        self.profile = profile
        self.hidden = hidden

    def __repr__(self):
        return (f"User(name={self.name!r}, "
                f"state={self.state!r}, "
                f"uid={self.uid!r}, "
                f"group={self.group!r}, "
                f"groups={self.groups!r}, "
                f"append={self.append!r}, "
                f"comment={self.comment!r}, "
                f"home={self.home!r}, "
                f"shell={self.shell!r}, "
                f"password={'***' if self.password else None!r}, "
                f"create_home={self.create_home!r}, "
                f"move_home={self.move_home!r}, "
                f"system={self.system!r}, "
                f"force={self.force!r}, "
                f"remove={self.remove!r}, "
                f"generate_ssh_key={self.generate_ssh_key!r}, "
                f"ssh_key_bits={self.ssh_key_bits!r}, "
                f"ssh_key_file={self.ssh_key_file!r}, "
                f"ssh_key_comment={self.ssh_key_comment!r}, "
                f"ssh_key_passphrase={'***' if self.ssh_key_passphrase else None!r}, "
                f"ssh_key_type={self.ssh_key_type!r}, "
                f"update_password={self.update_password!r}, "
                f"expires={self.expires!r}, "
                f"password_lock={self.password_lock!r}, "
                f"password_expire_max={self.password_expire_max!r}, "
                f"password_expire_min={self.password_expire_min!r}, "
                f"password_expire_warn={self.password_expire_warn!r}, "
                f"password_expire_account_disable={self.password_expire_account_disable!r}, "
                f"local={self.local!r}, "
                f"skeleton={self.skeleton!r}, "
                f"umask={self.umask!r}, "
                f"uid_min={self.uid_min!r}, "
                f"uid_max={self.uid_max!r}, "
                f"seuser={self.seuser!r}, "
                f"login_class={self.login_class!r}, "
                f"authorization={self.authorization!r}, "
                f"role={self.role!r}, "
                f"profile={self.profile!r}, "
                f"hidden={self.hidden!r})")

    def execute(self):
        # Build the command based on the parameters
        if self.state == "absent":
            cmd = f"userdel"
            if self.remove:
                cmd += " --remove"
            if self.force:
                cmd += " --force"
            cmd += f" {self.name}"
        else:
            cmd = f"useradd" if self.state == "present" else f"usermod"

            # Add common options
            if self.uid is not None:
                cmd += f" --uid {self.uid}"
            if self.group:
                cmd += f" --gid {self.group}"
            if self.groups:
                groups_str = ",".join(self.groups)
                if self.append:
                    cmd += f" --append --groups {groups_str}"
                else:
                    cmd += f" --groups {groups_str}"
            if self.comment:
                cmd += f" --comment '{self.comment}'"
            if self.home:
                cmd += f" --home {self.home}"
            if self.shell:
                cmd += f" --shell {self.shell}"
            if self.password and self.update_password == "always":
                cmd += f" --password '{self.password}'"
            if self.create_home:
                cmd += " --create-home"
            if self.move_home:
                cmd += " --move-home"
            if self.system:
                cmd += " --system"
            if self.expires is not None:
                cmd += f" --expiredate {self.expires}"
            if self.password_lock is not None:
                if self.password_lock:
                    cmd += " --lock"
                else:
                    cmd += " --unlock"

            # Add user name
            cmd += f" {self.name}"

            # Handle SSH key generation
            if self.generate_ssh_key:
                ssh_cmd = "ssh-keygen"
                if self.ssh_key_bits:
                    ssh_cmd += f" -b {self.ssh_key_bits}"
                if self.ssh_key_file:
                    ssh_cmd += f" -f {self.ssh_key_file}"
                if self.ssh_key_comment:
                    ssh_cmd += f" -C '{self.ssh_key_comment}'"
                if self.ssh_key_passphrase:
                    ssh_cmd += f" -P '{self.ssh_key_passphrase}'"
                else:
                    ssh_cmd += " -P ''"
                if self.ssh_key_type:
                    ssh_cmd += f" -t {self.ssh_key_type}"

                # Execute SSH key generation after user creation
                yield Command(ssh_cmd)

        # print(f"{self}")
        r = yield Command(cmd)
        r.changed = True


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command


class Fail:
    """
    A class to encapsulate the functionality of the Ansible fail module.
    This module fails the progress with a custom message.
    It can be useful for bailing out when a certain condition is met using when.

    Attributes:
        msg (str): The customized message used for failing execution.
        guard (bool): If `False` the commands will not be executed.

    **Examples:**

    .. code:: python

        # Fail with a custom message
        yield Fail("The system may not be provisioned according to the CMDB status.")

        # Fail with default message
        yield Fail()

    Usage:
        This class is designed to be used in a generator-based workflow where failure conditions are yielded for execution.

    Notes:
        - This module is also supported for Windows targets.
        - This module has a corresponding action plugin.
    """

    def __init__(self,
                 msg: str = "Failed as requested from task",
                 guard: bool = True):
        self.msg = msg
        self.guard = guard

    def __repr__(self):
        return (f"Fail(msg={self.msg!r}, "
                f"guard={self.guard!r})")

    def execute(self):
        # Construct the command to fail with custom message
        cmd = f"echo '{self.msg}' >&2 && exit 1"
        r = yield Command(cmd, guard=self.guard)
        r.changed = False
        # Mark the result as failed
        r.failed = True
        r.msg = self.msg


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import asyncio
from reemote.main import main
import re
import asyncssh
from asyncssh.sftp import SFTPAttrs

class Write_text_to_file:
    def execute(self):
        from reemote.operations.builtin.lineinfile import Lineinfile
        yield Lineinfile(
            line='new_config_value',
            path='/etc/config.conf',
            insertafter='^# Server configuration',  # Note: "Sever" not "Server"
            attrs = {'permissions': 0o755}  # Directly pass a dictionary
        )

if __name__ == "__main__":
    asyncio.run(main(Write_text_to_file))



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command


class Debconf:
    """
    A class to configure a .deb package using debconf-set-selections or query existing selections.

    This class allows users to pre-configure Debian packages by setting debconf database values,
    which is commonly used before package installation. It supports various value types including
    passwords, multiselect, boolean, etc.

    Attributes:
        name (str): Name of package to configure.
        question (str, optional): A debconf configuration setting/question.
        value (any, optional): Value to set the configuration to.
        vtype (str, optional): The type of the value supplied.
        unseen (bool): Do not set seen flag when pre-seeding.

    Examples:
        .. code:: python

            # Set default locale to fr_FR.UTF-8
            r = yield Debconf(
                name="locales",
                question="locales/default_environment_locale",
                value="fr_FR.UTF-8",
                vtype="select"
            )

            # Set to generate locales
            r = yield Debconf(
                name="locales",
                question="locales/locales_to_be_generated",
                value="en_US.UTF-8 UTF-8, fr_FR.UTF-8 UTF-8",
                vtype="multiselect"
            )

            # Accept oracle license
            r = yield Debconf(
                name="oracle-java7-installer",
                question="shared/accepted-oracle-license-v1-1",
                value=True,
                vtype="select"
            )

            # Query existing selections for a package
            r = yield Debconf(name="tzdata")

            # Pre-configure tripwire site passphrase
            r = yield Debconf(
                name="tripwire",
                question="tripwire/site-passphrase",
                value="my_secret_passphrase",
                vtype="password"
            )

    Notes:
        - This class requires the debconf and debconf-utils packages to be installed on the target system.
        - For sensitive data like passwords, ensure the task uses appropriate security measures.
        - This module only updates the debconf database; an additional step is needed to reconfigure the package.
        - Setting passwords may always show as changed due to debconf-get-selections masking passwords.

    """

    VALID_VTYPES = [
        "boolean", "error", "multiselect", "note",
        "password", "seen", "select", "string",
        "text", "title"
    ]

    def __init__(self,
                 name: str,
                 question: str = None,
                 value=None,
                 vtype: str = None,
                 unseen: bool = False):
        """
        Initialize a Debconf instance.

        Args:
            name (str): Name of package to configure (required).
            question (str, optional): A debconf configuration setting/question.
            value (any, optional): Value to set the configuration to.
            vtype (str, optional): The type of the value supplied.
            unseen (bool, optional): Do not set seen flag when pre-seeding. Defaults to False.
        """
        self.name = name
        self.question = question
        self.value = value
        self.vtype = vtype
        self.unseen = unseen

        # Validate vtype if provided
        if self.vtype and self.vtype not in self.VALID_VTYPES:
            raise ValueError(f"Invalid vtype '{self.vtype}'. Must be one of {self.VALID_VTYPES}")

    def __repr__(self):
        return (f"Debconf(name={self.name!r}, "
                f"question={self.question!r}, "
                f"value={self.value!r}, "
                f"vtype={self.vtype!r}, "
                f"unseen={self.unseen!r})")

    def _build_command(self) -> str:
        """
        Build the debconf-set-selections command string.

        Returns:
            str: The complete command string.
        """
        if not self.question or self.value is None or not self.vtype:
            # If missing required parameters, just check/show existing config
            return f"debconf-show {self.name}"

        # Format the value appropriately
        if isinstance(self.value, bool):
            formatted_value = "true" if self.value else "false"
        elif isinstance(self.value, list) and self.vtype == "multiselect":
            formatted_value = ", ".join(str(v) for v in self.value)
        else:
            formatted_value = str(self.value)

        # Escape special characters in value
        escaped_value = formatted_value.replace("'", "'\"'\"'")

        line = f"{self.name} {self.question} {self.vtype} '{escaped_value}'"
        cmd = f"echo '{line}' | debconf-set-selections"

        if self.unseen:
            cmd += " -u"

        return cmd

    def execute(self):
        """
        Execute the debconf configuration command.

        Yields:
            Command: A Command object for execution.
        """
        cmd = self._build_command()
        r = yield Command(cmd)
        r.changed = True


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command

class Stat:
    """
    A class to encapsulate the functionality of retrieving file or file system status,
    similar to the Linux/Unix stat command. This class allows users to specify a file path
    and various options to control what information is retrieved about the file.

    Attributes:
        path (str): The full path of the file/object to get the facts of.
        checksum_algorithm (str): Algorithm to determine checksum of file. Choices: "md5", "sha1", "sha224", "sha256", "sha384", "sha512".
        follow (bool): Whether to follow symlinks.
        get_attributes (bool): Get file attributes using lsattr tool if present.
        get_checksum (bool): Whether to return a checksum of the file.
        get_mime (bool): Use file magic and return data about the nature of the file.
        guard (bool): If `False` the commands will not be executed.

    **Examples:**

    .. code:: python

        # Obtain the stats of /etc/foo.conf
        r = yield Stat(path="/etc/foo.conf")
        # Check if file exists and belongs to 'root'
        if r.cp.stat.exists and r.cp.stat.pw_name != 'root':
            print("File ownership has changed")

        # Check if a path is a symlink
        r = yield Stat(path="/path/to/something")
        if r.cp.stat.islnk is not defined:
            print("Path doesn't exist")
        elif r.cp.stat.islnk:
            print("Path exists and is a symlink")
        else:
            print("Path exists and isn't a symlink")

        # Check if a path is a directory
        r = yield Stat(path="/path/to/something")
        if r.cp.stat.isdir is defined and r.cp.stat.isdir:
            print("Path exists and is a directory")

        # Do not calculate the checksum
        r = yield Stat(path="/path/to/myhugefile", get_checksum=False)

        # Use sha256 to calculate the checksum
        r = yield Stat(path="/path/to/something", checksum_algorithm="sha256")

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        - For Windows targets, use the ansible.windows.win_stat module instead.
        - Commands are constructed based on the provided parameters.
    """
    
    def __init__(self,
                 path: str,
                 checksum_algorithm: str = "sha1",
                 follow: bool = False,
                 get_attributes: bool = True,
                 get_checksum: bool = True,
                 get_mime: bool = True,
                 guard: bool = True):
        
        self.path = path
        self.checksum_algorithm = checksum_algorithm
        self.follow = follow
        self.get_attributes = get_attributes
        self.get_checksum = get_checksum
        self.get_mime = get_mime
        self.guard = guard

    def __repr__(self):
        return (f"Stat(path={self.path!r}, "
                f"checksum_algorithm={self.checksum_algorithm!r}, "
                f"follow={self.follow!r}, "
                f"get_attributes={self.get_attributes!r}, "
                f"get_checksum={self.get_checksum!r}, "
                f"get_mime={self.get_mime!r}, "
                f"guard={self.guard!r})")

    def execute(self):
        # Build the command arguments
        args = []
        
        # Add path argument
        args.append(f"path={self.path}")
        
        # Add optional arguments
        if self.checksum_algorithm != "sha1":  # sha1 is default
            args.append(f"checksum_algorithm={self.checksum_algorithm}")
            
        if self.follow:
            args.append("follow=yes")
            
        if not self.get_attributes:  # default is True
            args.append("get_attributes=no")
            
        if not self.get_checksum:  # default is True
            args.append("get_checksum=no")
            
        if not self.get_mime:  # default is True
            args.append("get_mime=no")
        
        # Construct the command
        cmd_args = ",".join(args)
        cmd = f"stat {cmd_args}"
        
        # Execute the command
        r = yield Command(cmd, guard=self.guard)
        r.changed = False  # stat commands don't change system state


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.command import Command


class AptRepository:
    """
    A class to manage APT repositories on Debian-based systems.
    This class allows adding or removing APT repositories similar to the ansible.builtin.apt_repository module.

    Attributes:
        repo (str): A source string for the repository.
        state (str): Desired state of the repository ('present' or 'absent').
        codename (str): Override the distribution codename for PPA repositories.
        filename (str): Sets the name of the source list file in sources.list.d.
        install_python_apt (bool): Whether to automatically install python3-apt.
        mode (str): The octal mode for newly created files in sources.list.d.
        update_cache (bool): Run apt-get update when a change occurs.
        update_cache_retries (int): Amount of retries if the cache update fails.
        update_cache_retry_max_delay (int): Max delay for exponential backoff.
        validate_certs (bool): If false, SSL certificates will not be validated.

    **Examples:**

    .. code:: python

        # Add a repository
        r = yield AptRepository(
            repo="deb http://archive.canonical.com/ubuntu hardy partner",
            state="present"
        )
        print(r.cp.stdout)

        # Remove a repository
        r = yield AptRepository(
            repo="deb http://archive.canonical.com/ubuntu hardy partner",
            state="absent"
        )
        print(r.cp.stdout)

        # Add repository with custom filename
        r = yield AptRepository(
            repo="deb http://dl.google.com/linux/chrome/deb/ stable main",
            state="present",
            filename="google-chrome"
        )
        print(r.cp.stdout)

    Usage:
        This class is designed to be used in a generator-based workflow where commands are yielded for execution.

    Notes:
        - Requires python3-apt and apt-key or gpg on the target system
        - Supports Debian Squeeze (version 6) and its successors
        - Works with both regular repositories and PPAs
    """

    def __init__(self,
                 repo: str,
                 state: str = "present",
                 codename: str = None,
                 filename: str = None,
                 install_python_apt: bool = True,
                 mode: str = None,
                 update_cache: bool = True,
                 update_cache_retries: int = 5,
                 update_cache_retry_max_delay: int = 12,
                 validate_certs: bool = True):

        self.repo = repo
        self.state = state
        self.codename = codename
        self.filename = filename
        self.install_python_apt = install_python_apt
        self.mode = mode
        self.update_cache = update_cache
        self.update_cache_retries = update_cache_retries
        self.update_cache_retry_max_delay = update_cache_retry_max_delay
        self.validate_certs = validate_certs

    def __repr__(self):
        return (f"AptRepository(repo={self.repo!r}, "
                f"state={self.state!r}, "
                f"codename={self.codename!r}, "
                f"filename={self.filename!r}, "
                f"install_python_apt={self.install_python_apt!r}, "
                f"mode={self.mode!r}, "
                f"update_cache={self.update_cache!r}, "
                f"update_cache_retries={self.update_cache_retries!r}, "
                f"update_cache_retry_max_delay={self.update_cache_retry_max_delay!r}, "
                f"validate_certs={self.validate_certs!r})")

    def execute(self):
        # Build the apt_repository command
        cmd_parts = ["apt_repository"]

        # Add required repo parameter
        cmd_parts.append(f"repo='{self.repo}'")

        # Add state parameter
        cmd_parts.append(f"state={self.state}")

        # Add optional parameters if specified
        if self.codename is not None:
            cmd_parts.append(f"codename={self.codename}")

        if self.filename is not None:
            cmd_parts.append(f"filename={self.filename}")

        if not self.install_python_apt:
            cmd_parts.append("install_python_apt=false")

        if self.mode is not None:
            cmd_parts.append(f"mode={self.mode}")

        if not self.update_cache:
            cmd_parts.append("update_cache=false")

        cmd_parts.append(f"update_cache_retries={self.update_cache_retries}")
        cmd_parts.append(f"update_cache_retry_max_delay={self.update_cache_retry_max_delay}")

        if not self.validate_certs:
            cmd_parts.append("validate_certs=false")

        # Construct the full command
        full_cmd = " ".join(cmd_parts)

        # Execute the command
        r = yield Command(full_cmd, guard=True, sudo=True)
        r.changed = True


# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from nicegui import ui

from reemote.gui.functions.download_file import download_file
from reemote.gui.local_file_picker.local_file_picker import local_file_picker
from reemote.gui.functions.upload_file import upload_file

class File_picker:
    def __init__(self):
        self.paths= "/"

async def pick_file(fp: File_picker) -> None:
    result = await local_file_picker('~', multiple=True)
    ui.notify(f'You chose {result}')
    fp.paths=result

def file_manager(tabs, inventory, versions, manager, stdout_report, execution_report, file_path):

    fp = File_picker()

    with ui.tab_panels(tabs, value='File Manager').classes('w-full'):
        with ui.tab_panel('File Manager'):


            with ui.row():
                ui.input(label='Server builtin path').bind_value(file_path, 'path')
                ui.markdown("""
                This is the name of the builtin on the servers.  
                """)
            with ui.row():
                ui.button('Download File', on_click=lambda: download_file(inventory, file_path, stdout_report, execution_report))
                ui.markdown("""
                Download the builtin content from the first server in the inventory.  
                """)
            with ui.row():
                ui.button('Choose Files', on_click=lambda: pick_file(fp), icon='folder')
                ui.markdown("""
                Choose files on the local host.  
                """)
            with ui.row():
                ui.button('Upload Files', on_click=lambda: upload_file(fp, inventory, file_path, stdout_report, execution_report))
                ui.markdown("""
                Upload a builtin from the local host to all the servers in the inventory.  
                """)




# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
from reemote.operations.server.shell import Shell
from reemote.execute import execute

def parse_apt_list_installed(output):
    """Parses the output of 'apt list --installed' into a list of dictionaries.

    This helper function processes the raw string output from the
    `apt list --installed` command. It iterates through each line,
    skipping the "Listing..." header and any empty lines. For each valid
    package line, it accurately extracts the package name and its
    corresponding version number.

    Args:
        output (str): The raw string output from the `apt list --installed`
            command.

    Returns:
        list[dict]: A list of dictionaries, where each dictionary
            represents an installed package and contains 'name' and 'version'
            keys. Example: `[{'name': 'zlib1g', 'version': '1:1.2.11.dfsg-2'}]`.
    """
    lines = output.strip().split('\n')
    packages = []

    for line in lines:
        line = line.strip()
        if not line or line.startswith('Listing...'):
            continue

        # Split package name from the rest using first '/'
        if '/' not in line:
            continue

        name_part, rest = line.split('/', 1)
        name = name_part.strip()

        # Find the first space — version starts right after it
        space_index = rest.find(' ')
        if space_index == -1:
            continue

        # Extract everything after the first space
        after_space = rest[space_index + 1:]

        # Version is everything until the next space or '['
        version = after_space.split(' ', 1)[0].split('[', 1)[0].rstrip(',')

        packages.append({"name": name, "version": version})

    return packages

class Get_packages:
    """A reemote fact that gathers installed APT packages from a Debian-based system.

    When executed by the reemote framework, this operation runs the
    `apt list --installed` command on the target server. It then parses the
    command's output to create a structured list of all installed packages,
    including their names and versions.

    The final result is a list of dictionaries, which is attached to the
    `stdout` attribute of the completed process object. This operation is
    read-only and will always have `changed=False`.

    **Examples:**

    .. code:: python

        yield Get_packages()


    .. code:: bash

        reemote -i ~/reemote/inventory-proxmox-debian.py -s reemote/facts/apt/get_packages.py -c Get_packages
    """
    def execute(self):
        from reemote.operations.server.shell import Shell
        r = yield Shell("apt list --installed")
        r.cp.stdout = parse_apt_list_installed(r.cp.stdout)
        r.changed = False
        # print(r.cp.stdout)



# Copyright (c) 2025 Kim Jarvis TPF Software Services S.A. kim.jarvis@tpfsystems.com 
# This software is licensed under the MIT License. See the LICENSE file for details.
#
import sys


def make_inventory(
        inventory_filename: str,
        image: str,
        vm: str,
        name: str,
        user: str,
        user_password: str,
        root_password: str,
        ip_address: str
) -> None:
    """Generates a Python-based inventory builtin for a virtual machine.

    This function creates a Python script that serves as a configuration
    inventory. The generated script contains a single function, `inventory()`,
    which returns the connection and privilege escalation details for a
    specific VM host. The content of the builtin is dynamically generated
    using the provided arguments.

    The resulting inventory builtin is structured to be compatible with tools
    that expect a Python-based configuration module.

    Args:
        inventory_filename (str): The path and name for the output inventory builtin.
        image (str): A description of the VM image (e.g., "Ubuntu 22.04").
        vm (str): A name or identifier for the virtual machine.
        name (str): The full name of the user, used for generating comments.
        user (str): The username for SSH and sudo access.
        user_password (str): The password for the regular user account.
        root_password (str): The password for the root account, used for `su`.
        ip_address (str): The IP address of the target virtual machine.

    Returns:
        None

    Raises:
        IOError: If an error occurs while writing to the inventory builtin.
    """
    # Create the inventory builtin content
    inventory_content = f"""from typing import List, Tuple, Dict, Any

def inventory() -> List[Tuple[Dict[str, Any], Dict[str, str]]]:
    return [
        (
            {{
                'host': '{ip_address}',  # {image} {vm}
                'username': '{user}',  # {name}
                'password': '{user_password}',  # Password
            }},
            {{
                'sudo_user': '{user}',  # Sudo user
                'sudo_password': '{user_password}',  # Password
                'su_user': 'root',  # su user
                'su_password': '{root_password}'  # su Password
            }}
        )
    ]"""

    # Write the inventory builtin
    try:
        with open(inventory_filename, 'w') as f:
            f.write(inventory_content)
        print(f"Inventory builtin '{inventory_filename}' created successfully.")
    except IOError as e:
        print(f"Error writing inventory builtin '{inventory_filename}': {e}", file=sys.stderr)
        raise
